# -*- coding: utf-8 -*-
"""
  Copyright 2022 Mitchell Isaac Parker

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.

"""

import os
import pandas as pd
import xml.etree.ElementTree as ET
import gzip
from tqdm import tqdm
import concurrent.futures
from Bio.PDB import Select
import pymol2

from ..functions.coord import (
    load_coord,
    load_cif_dict,
    save_coord,
    build_pdb_dict,
    build_pdb_code_lst,
    get_reschainid,
    get_resid,
    get_atomid,
    get_modelid,
    get_neighbors,
    get_chainid,
    get_chain_cont,
    get_resnum,
    get_resname,
    is_aa,
    is_het,
    is_wat,
    get_residue_cont,
)
from ..functions.path import (
    path_exists,
    get_renum_path,
    get_dir_path,
    save_table,
    save_json,
    get_core_path,
    get_rcsb_path,
    get_sifts_path,
    search_dir,
    append_path,
    delete_path,
    core_str,
    sifts_str,
    rcsb_str,
    renum_str,
    rcsb_assembly_str,
    renum_assembly_str,
)
from ..functions.lig import lig_lst_dict
from ..functions.lst import lst_to_str, res_to_str, type_lst, lst_unique, res_to_lst
from ..functions.table import get_str_num, merge_dicts, merge_tables, lst_col
from ..functions.col import (
    pdb_id_col,
    pdb_code_col,
    chainid_col,
    modelid_col,
    bound_lig_col,
    bound_prot_chainid_col,
    range_col,
    date_col,
    pmid_col,
    core_path_col,
    rcsb_path_col,
    renum_path_col,
    sifts_path_col,
    rcsb_assembly_path_col,
    renum_assembly_path_col,
    bio_lig_col,
    ion_lig_col,
    chem_lig_col,
    mod_lig_col,
    mem_lig_col,
    pharm_lig_col,
)

max_prot_cb_dist = 12.0
max_prot_atomid_dist = 5.0
min_prot_cb_cont = 12
min_prot_cb_atomid_cont = 1
min_prot_atomid_cont = 5

max_lig_resid_dist = 4
min_lig_resid_cont = 5


class ChainSelect(Select):
    def __init__(self, sele_dict):
        self.sele_dict = sele_dict

    def accept_residue(self, residue):

        return self.sele_dict[get_reschainid(residue)][get_resid(residue)]

    def accept_atom(self, atom):

        if get_atomid(atom)[0] == "H":
            return 0
        else:
            return 1


def run_pdb_renum(
    pdb_code_lst,
    renum_script_path,
    rcsb_dir=None,
    sifts_dir=None,
    renum_dir=None,
    num_cpu=1,
):

    renum_pdb_code_lst = [
        x
        for x in pdb_code_lst
        if not path_exists(get_renum_path(x, dir_path=renum_dir))
    ]

    if len(renum_pdb_code_lst) > 0:
        pdb_code_str = lst_to_str(
            renum_pdb_code_lst,
            join_txt=" ",
        )

        cmd_lst = [
            f"python {renum_script_path}",
            f"-rfla {pdb_code_str}",
            "-mmCIF",
            f"-sipm {get_dir_path(dir_str=rcsb_str, dir_path=rcsb_dir)}",
            f"-sips {get_dir_path(dir_str=sifts_str,dir_path=sifts_dir)}",
            f"-sopm {get_dir_path(dir_str=renum_str, dir_path=renum_dir)}",
            "-offz",
            f"-nproc {num_cpu}",
        ]

        cmd_str = lst_to_str(cmd_lst, join_txt=" ")

        os.system(cmd_str)

        assembly_cmd_lst = [
            f"python {renum_script_path}",
            f"-rfla {pdb_code_str}",
            "-mmCIF_assembly",
            f"-sipma {get_dir_path(dir_str=rcsb_assembly_str, dir_path=rcsb_dir)}",
            f"-sips {get_dir_path(dir_str=sifts_str,dir_path=sifts_dir)}",
            f"-sopma {get_dir_path(dir_str=renum_assembly_str, dir_path=renum_dir)}",
            "-offz",
            f"-nproc {num_cpu}",
        ]

        assembly_cmd_str = lst_to_str(assembly_cmd_lst, join_txt=" ")

        os.system(assembly_cmd_str)


def get_sifts_dict(sifts_path):

    tree = ET.parse(gzip.open(sifts_path, "rt"))
    root = tree.getroot()
    base = "{http://www.ebi.ac.uk/pdbe/docs/sifts/eFamily.xsd}"

    sifts_dict = dict()
    uniprot_dict = dict()

    pdb_chainid_lst = list()
    uniprot_chainid_lst = list()

    for entity in root:
        if entity.tag == (str(base) + "entity"):
            for segment in entity:
                if segment.tag == (str(base) + "segment"):
                    for listResidue in segment:
                        if listResidue.tag == (str(base) + "listResidue"):
                            for residue in listResidue:
                                if residue.tag == (str(base) + "residue"):
                                    if residue.get("dbSource") == "PDBe":

                                        pdbe_resid = residue.get("dbResNum")

                                        pdb_resid = "null"
                                        uniprot_resid = "null"

                                        no_uniprot = True

                                        for crossRefDb in residue:
                                            if crossRefDb.get("dbSource") == "PDB":
                                                pdb_code = crossRefDb.get(
                                                    "dbAccessionId"
                                                )
                                                chainid = crossRefDb.get("dbChainId")

                                                if chainid not in pdb_chainid_lst:
                                                    pdb_chainid_lst.append(chainid)

                                                pdb_resid = crossRefDb.get("dbResNum")

                                            if crossRefDb.get("dbSource") == "UniProt":

                                                if chainid not in uniprot_chainid_lst:
                                                    uniprot_chainid_lst.append(chainid)

                                                no_uniprot = False
                                                uniprot_resid = crossRefDb.get(
                                                    "dbResNum"
                                                )

                                        if no_uniprot:
                                            uniprot_resid = str(
                                                get_str_num(pdbe_resid) + 50000
                                            )

                                            if chainid not in list(uniprot_dict.keys()):
                                                uniprot_dict[chainid] = dict()

                                            uniprot_dict[chainid][
                                                uniprot_resid
                                            ] = pdb_resid

                                        if (
                                            pdb_resid != "null"
                                            and uniprot_resid != "null"
                                        ):
                                            if pdb_code not in list(sifts_dict.keys()):
                                                sifts_dict[pdb_code] = dict()

                                            if chainid not in list(
                                                sifts_dict[pdb_code].keys()
                                            ):
                                                sifts_dict[pdb_code][chainid] = dict()

                                            sifts_dict[pdb_code][chainid][
                                                pdb_resid
                                            ] = uniprot_resid

    for chainid in pdb_chainid_lst:
        if chainid not in uniprot_chainid_lst:
            if chainid in list(uniprot_dict.keys()):
                for pdb_resid in list(sifts_dict[pdb_code][chainid].keys()):
                    uniprot_resid = sifts_dict[pdb_code][chainid][pdb_resid]
                    if int(uniprot_resid) > 50000:
                        sifts_dict[pdb_code][chainid][pdb_resid] = uniprot_dict[
                            chainid
                        ][uniprot_resid]

    return sifts_dict


def build_sifts_map(sifts_path_lst, sifts_json_path, num_cpu=1):

    sifts_dict = {}

    with concurrent.futures.ProcessPoolExecutor(max_workers=num_cpu) as executor:

        job_lst = [
            executor.submit(get_sifts_dict, sifts_path) for sifts_path in sifts_path_lst
        ]

        for job in tqdm(
            concurrent.futures.as_completed(job_lst),
            desc="Preparing SIFTS numbering",
            total=len(job_lst),
            miniters=1,
            position=0,
            leave=True,
        ):

            sifts_dict = merge_dicts([sifts_dict, job.result()])

    save_json(sifts_json_path, sifts_dict)


def build_sele_dict(
    structure,
    curr_chainid,
    chainid_lst,
    resid_cont_dict=None,
    bound_chainid_lst=None,
):

    sele_dict = dict()
    sele_df = pd.DataFrame()

    is_mem = False

    bound_lig_lst = list()
    bound_prot_chainid_lst = list()

    for model in structure:

        modelid = get_modelid(model)

        sele_dict[modelid] = dict()

        curr_chain = model[curr_chainid]

        neighbors = get_neighbors(curr_chain)

        resnum_lst = list()

        for chain in model:

            chainid = get_chainid(chain)

            sele_dict[modelid][chainid] = dict()

            for residue in chain:

                resid = get_resid(residue)
                sele = 0

                if is_aa(residue):

                    if chainid == curr_chainid:
                        resnum = get_resnum(residue)
                        if resnum < 50000:
                            resnum_lst.append(resnum)
                        sele = 1

                elif is_het(residue):

                    resname = get_resname(residue)

                    is_pharm = True
                    for lig_col, lig_lst in lig_lst_dict.items():
                        if resname in lig_lst:
                            is_pharm = False
                            break
                    if is_pharm:
                        lig_col = pharm_lig_col

                    if lig_col == mem_lig_col:
                        sele = 1
                        is_mem = True
                    elif chainid == curr_chainid and lig_col in [
                        bio_lig_col,
                        ion_lig_col,
                        chem_lig_col,
                        mod_lig_col,
                    ]:
                        sele = 1
                    elif lig_col == pharm_lig_col:
                        resid_residue_cont = lst_unique(
                            [
                                get_resnum(x)
                                for x in get_residue_cont(
                                    neighbors,
                                    residue,
                                    max_dist=max_lig_resid_dist,
                                    level="R",
                                )
                                if is_aa(x) and get_reschainid(x) == chainid
                            ]
                        )

                        if resid_cont_dict is not None:
                            if lig_col in list(resid_cont_dict.keys()):
                                resid_residue_cont = [
                                    x
                                    for x in resid_residue_cont
                                    if get_resnum(x) in resid_cont_dict[lig_col]
                                ]

                        resid_residue_cont = len(resid_residue_cont)

                        if resid_residue_cont >= min_lig_resid_cont:
                            sele = 1

                    if sele == 1:
                        bound_lig_lst.append(resname)

                elif is_wat(residue):
                    if chainid == curr_chainid:
                        sele = 1

                sele_dict[modelid][chainid][resid] = sele

            if chainid not in chainid_lst:
                sele = 0

                check_chainid = True
                if bound_chainid_lst is not None:
                    if chainid not in bound_chainid_lst:
                        check_chainid = False

                if is_mem:
                    sele = 1
                    check_chainid = False

                if check_chainid:

                    atomid_chain_cont = [
                        x
                        for x in get_chain_cont(
                            neighbors,
                            chain,
                            max_dist=max_prot_atomid_dist,
                        )
                        if is_aa(x.get_parent())
                        and get_reschainid(x.get_parent()) == curr_chainid
                    ]

                    cb_chain_cont = [
                        x
                        for x in get_chain_cont(
                            neighbors,
                            chain,
                            max_dist=max_prot_cb_dist,
                        )
                        if is_aa(x.get_parent())
                        and get_reschainid(x.get_parent()) == curr_chainid
                        and get_atomid(x) == "CB"
                    ]

                    if resid_cont_dict is not None:
                        if bound_prot_chainid_col in list(resid_cont_dict.keys()):
                            atomid_chain_cont = [
                                x
                                for x in atomid_chain_cont
                                if get_resnum(x.get_parent())
                                in resid_cont_dict[bound_prot_chainid_col]
                            ]
                            cb_chain_cont = [
                                x
                                for x in cb_chain_cont
                                if get_resnum(x.get_parent())
                                in resid_cont_dict[bound_prot_chainid_col]
                            ]

                    atomid_chain_cont = len(atomid_chain_cont)
                    cb_chain_cont = len(cb_chain_cont)

                    if (
                        cb_chain_cont >= min_prot_cb_cont
                        and atomid_chain_cont >= min_prot_cb_atomid_cont
                    ) or atomid_chain_cont >= min_prot_atomid_cont:

                        sele = 1

                if sele == 1:
                    bound_prot_chainid_lst.append(chainid)

                for residue in chain:
                    resid = get_resid(residue)
                    sele_dict[modelid][chainid][resid] = sele

        sele_df.at[modelid, modelid_col] = modelid
        sele_df.at[modelid, bound_lig_col] = lst_to_str(bound_lig_lst)
        sele_df.at[modelid, bound_prot_chainid_col] = lst_to_str(bound_prot_chainid_lst)
        sele_df.at[modelid, range_col] = res_to_str(resnum_lst)

    return sele_dict, sele_df


def isolate_chains(
    pdb_code,
    pdb_dict,
    core_dir=None,
    rcsb_dir=None,
    sifts_dir=None,
    renum_dir=None,
    resid_cont_dict=None,
    add_h=False,
    add_h_his=False,
    bound_chainid_dict=None,
    all_models=False,
    update_coords=False,
):

    rcsb_path = get_rcsb_path(pdb_code, dir_path=rcsb_dir)
    renum_path = get_renum_path(pdb_code, dir_path=renum_dir)
    sifts_path = get_sifts_path(pdb_code, dir_path=sifts_dir)

    try:
        chainid_lst = pdb_dict[pdb_code]

        structure = load_coord(renum_path)
        renum_dict = load_cif_dict(renum_path)

        deposit_date = renum_dict["_pdbx_database_status.recvd_initial_deposition_date"][0]

        if not all_models:
            structure = type_lst(structure[0])

        chain_df = pd.DataFrame()

        for chainid in chainid_lst:

            pdb_id = f"{pdb_code}{chainid}"

            bound_chainid_lst = None
            if bound_chainid_dict is not None:
                if pdb_id in list(bound_chainid_dict.keys()):
                    bound_chainid_lst = bound_chainid_dict[pdb_id]

            rcsb_assembly_path = "None"
            renum_assembly_path = "None"

            if bound_chainid_lst is None:

                renum_assembly_dir = get_dir_path(
                    dir_str=renum_assembly_str, dir_path=renum_dir
                )
                rcsb_assembly_dir = get_dir_path(
                    dir_str=rcsb_assembly_str, dir_path=rcsb_dir
                )

                for renum_assembly_file in search_dir(renum_assembly_dir, pdb_code):

                    renum_assembly_path = f"{renum_assembly_dir}/{renum_assembly_file}"
                    rcsb_assembly_path = f"{rcsb_assembly_dir}/{renum_assembly_file.replace('_renum.cif','.cif.gz')}"

                    assembly_dict = load_cif_dict(renum_assembly_path)

                    assembly_chainid_lst = lst_unique(
                        assembly_dict["_pdbe_chain_remapping.orig_auth_asym_id"]
                    )

                    if chainid in assembly_chainid_lst:
                        if (
                            len(
                                [
                                    x
                                    for x in assembly_chainid_lst
                                    if x in chainid_lst and x != chainid
                                ]
                            )
                            == 0
                        ):
                            bound_chainid_lst = [
                                x for x in assembly_chainid_lst if x != chainid
                            ]

            sele_dict, sele_df = build_sele_dict(
                structure,
                chainid,
                chainid_lst,
                resid_cont_dict=resid_cont_dict,
                bound_chainid_lst=bound_chainid_lst,
            )

            sele_df[pdb_code_col] = pdb_code
            sele_df[chainid_col] = chainid
            sele_df[rcsb_path_col] = rcsb_path
            sele_df[renum_path_col] = renum_path
            sele_df[sifts_path_col] = sifts_path
            sele_df[rcsb_assembly_path_col] = rcsb_assembly_path
            sele_df[renum_assembly_path_col] = renum_assembly_path

            for modelid in list(sele_dict.keys()):

                modelid_str = modelid
                if not all_models:
                    modelid_str = None

                cif_path = get_core_path(
                    pdb_code, chainid, modelid=modelid_str, dir_path=core_dir
                )
                pdb_path = get_core_path(
                    pdb_code,
                    chainid,
                    modelid=modelid_str,
                    return_pdb=True,
                    dir_path=core_dir,
                )

                update_cif = True
                if not update_coords:
                    if path_exists(cif_path):
                        update_cif = False

                update_pdb = True
                if not update_coords:
                    if path_exists(cif_path):
                        update_pdb = False

                sele_df.at[int(modelid), core_path_col] = cif_path

                if update_cif:
                    chain_sele = ChainSelect(sele_dict[modelid])
                    save_coord(cif_path, structure[int(modelid)], sele=chain_sele)

                if update_pdb:
                    pymol_obj = f"{pdb_code}{chainid}"

                    if all_models:
                        pymol_obj += str(modelid)

                    with pymol2.PyMOL() as pymol:
                        cmd = pymol.cmd
                        cmd.load(cif_path, pymol_obj)
                        cmd.save(pdb_path, pymol_obj)

                if add_h:
                    pdb_h_path = get_core_path(
                        pdb_code,
                        chainid,
                        modelid=modelid_str,
                        return_pdb=True,
                        add_h=True,
                        dir_path=core_dir,
                    )

                    update_pdb_h = True
                    if not update_coords:
                        if path_exists(pdb_h_path):
                            update_pdb_h = False

                    if update_pdb_h:
                        cmd_lst = ["reduce", pdb_path, ">", pdb_h_path, "-NOHETh"]

                        if add_h_his:
                            cmd_lst.append("-HIS")

                        os.system(lst_to_str(cmd_lst, join_txt=" "))

                sele_df[pdb_id_col] = sele_df[pdb_code_col].map(str) + sele_df[
                    chainid_col
                ].map(str)

                if all_models:
                    sele_df[pdb_id_col] += sele_df[modelid_col].map(str)

                sele_df[modelid_col] = 0

                sele_df[date_col] = deposit_date

                chain_df = pd.concat([chain_df, sele_df], sort=False)
    except:
        print(f"Error loading PDB: {pdb_code}. Removing entry.")
        chain_df = pd.DataFrame()

    return chain_df


def run_pdb_chain(
    pdb_ids,
    core_dir=None,
    rcsb_dir=None,
    sifts_dir=None,
    renum_dir=None,
    resid_cont_dict=None,
    add_h=False,
    add_h_his=False,
    bound_chainid_dict=None,
    all_models=False,
    update_coords=False,
    num_cpu=1,
):

    append_path(get_dir_path(dir_str=core_str, dir_path=core_dir))

    pdb_id_lst = type_lst(pdb_ids)

    pdb_dict = build_pdb_dict(pdb_id_lst)

    if resid_cont_dict is not None:
        for key, val in resid_cont_dict.items():
            resid_cont_dict[key] = res_to_lst(val)

    df = pd.DataFrame()

    if num_cpu == 1:
        for pdb_code in tqdm(
            list(pdb_dict.keys()),
            desc="Isolating PDB chains",
            position=0,
            leave=True,
        ):
            df = pd.concat(
                [
                    df,
                    isolate_chains(
                        pdb_code,
                        pdb_dict,
                        core_dir=core_dir,
                        rcsb_dir=rcsb_dir,
                        sifts_dir=sifts_dir,
                        renum_dir=renum_dir,
                        resid_cont_dict=resid_cont_dict,
                        add_h=add_h,
                        add_h_his=add_h_his,
                        bound_chainid_dict=bound_chainid_dict,
                        all_models=all_models,
                        update_coords=update_coords,
                    ),
                ],
                sort=False,
            )
    else:
        with concurrent.futures.ProcessPoolExecutor(max_workers=num_cpu) as executor:
            job_lst = [
                executor.submit(
                    isolate_chains,
                    pdb_code,
                    pdb_dict,
                    core_dir=core_dir,
                    rcsb_dir=rcsb_dir,
                    sifts_dir=sifts_dir,
                    renum_dir=renum_dir,
                    resid_cont_dict=resid_cont_dict,
                    add_h=add_h,
                    add_h_his=add_h_his,
                    bound_chainid_dict=bound_chainid_dict,
                    all_models=all_models,
                    update_coords=update_coords,
                )
                for pdb_code in list(pdb_dict.keys())
            ]

            for job in tqdm(
                concurrent.futures.as_completed(job_lst),
                desc="Isolating PDB chains",
                total=len(job_lst),
                miniters=1,
                position=0,
                leave=True,
            ):

                df = pd.concat([df, job.result()], sort=False)

    df = df.reset_index(drop=True)

    return df


def prep_coord(
    pdb_id_lst,
    renum_script_path,
    coord_table_path,
    core_dir=None,
    rcsb_dir=None,
    sifts_dir=None,
    renum_dir=None,
    sifts_json_path=None,
    resid_cont_dict=None,
    add_h=False,
    add_h_his=False,
    bound_chainid_dict=None,
    all_models=False,
    data=None,
    update_coords=False,
    num_cpu=1,
):
    pdb_code_lst = build_pdb_code_lst(pdb_id_lst)

    run_pdb_renum(
        pdb_code_lst,
        renum_script_path,
        rcsb_dir=rcsb_dir,
        sifts_dir=sifts_dir,
        renum_dir=renum_dir,
        num_cpu=num_cpu,
    )

    delete_path("log_corrected.tsv")
    delete_path("log_translator.tsv")

    if sifts_json_path is not None:
        
        sifts_path_lst = [
            get_sifts_path(pdb_code, dir_path=sifts_dir) for pdb_code in pdb_code_lst
        ]

        build_sifts_map(sifts_path_lst, sifts_json_path, num_cpu=num_cpu)

    df = run_pdb_chain(
        pdb_id_lst,
        core_dir=core_dir,
        rcsb_dir=rcsb_dir,
        sifts_dir=sifts_dir,
        renum_dir=renum_dir,
        resid_cont_dict=resid_cont_dict,
        add_h=add_h,
        add_h_his=add_h_his,
        bound_chainid_dict=bound_chainid_dict,
        all_models=all_models,
        update_coords=update_coords,
        num_cpu=num_cpu,
    )

    if len(df) > 0:
        if data is not None:
            df_col_lst = list(data.columns)

            for col in [
                pdb_id_col,
                bound_lig_col,
                bound_prot_chainid_col,
                range_col,
                date_col,
                pmid_col,
            ]:
                if col in df_col_lst:
                    del data[col]

            df = merge_tables(df, data)

        save_table(coord_table_path, df)

        if len(df) > 0:
            print(f"Prepared {len(lst_col(df, pdb_id_col, unique=True))}/{len(pdb_id_lst)} chains from {len(lst_col(df, pdb_code_col,unique=True))}/{len(pdb_code_lst)} entries!")

    print("Prepared coordinates!")


# -*- coding: utf-8 -*-
"""
  Copyright 2022 Mitchell Isaac Parker

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.

"""

y32_col = "Y32"
y71_col = "Y71"
sw1_col = "SW1"
sw2_col = "SW2"

id_col = "id"

core_path_col = "core_path"
rcsb_path_col = "rcsb_path"
renum_path_col = "renum_path"
rcsb_assembly_path_col = "rcsb_assembly_path"
renum_assembly_path_col = "renum_assembly_path"
sifts_path_col = "sifts_path"
edia_path_col = "edia_path"
interf_path_col = "interface_path"
pocket_path_col = "pocket_path"

modelid_col = "modelid"
chainid_col = "chainid"
resid_col = "resid"
resname_col = "resname"
atomid_col = "atomid"

pdb_id_col = "pdb_id"
pdb_code_col = "pdb_code"
swiss_id_col = "swiss_id"
uniprot_id_col = "uniprot_id"
pfam_col = "pfam"
prot_col = "prot_name"
method_col = "method"
resolution_col = "resolution"
r_factor_col = "r_factor"
seq_col = "sequence"
len_col = "length"
range_col = "range"
date_col = "deposit_data"
pmid_col = "pmid"

len_a_col = "length_a"
len_b_col = "length_b"
len_c_col = "length_c"
ang_a_col = "angle_alpha"
ang_b_col = "angle_beat"
ang_g_col = "angle_gamma"
space_col = "space_group"
cf_col = "crystal_form"

mut_status_col = "mutation_status"
mut_pos_col = "mutation_position"

interf_id_col = "interface_id"
interf_col = "interface"
interf_area_col = "interface_area"
cb_cont_col = "cb_contacts"
atomid_cont_col = "atomid_contacts"
total_cb_cont_col = "total_cb_contacts"
total_atomid_cont_col = "total_atomid_contacts"
interf_cont_col = "interface_contacts"
cb_dist_col = "cb_distance"
iso_col = "isologous"

pocket_id_col = "pocket_id"
pocket_col = "pocket"
pocket_site_col = "pocket_site"
pocket_volume_col = "pocket_volume"
pocket_score_col = "pocket_score"
pocket_status_col = "pocket_status"
pocket_type_col = "pocket_type"
pocket_cont_col = "pocket_contacts"

bio_lig_col = "biological_ligand"
ion_lig_col = "ion_ligand"
pharm_lig_col = "pharmacological_ligand"
chem_lig_col = "chemical_ligand"
mod_lig_col = "modification_ligand"
mem_lig_col = "membrane_ligand"
pocket_lig_col = "pocket_ligand"

gene_class_col = "gene_class"
nuc_class_col = "nucleotide_class"
mut_class_col = "mutation_class"
pharm_class_col = "pharmacological_class"
match_class_col = "match_class"
prot_class_col = "protein_class"
interf_class_col = "interface_class"
pocket_class_col = "pocket_class"

bound_lig_col = "bound_ligand"
bound_prot_col = "bound_protein"
bound_prot_swiss_id_col = "bound_protein_swiss_id"

bound_lig_cont_col = 'bound_ligand_contacts'
bound_prot_cont_col = 'bound_protein_contacts'

pharm_lig_site_col = "pharmacological_ligand_site"
pharm_lig_match_col = "pharmarmacological_ligand_match"
pharm_lig_smiles_col = "pharmacological_ligand_smiles"

bound_prot_pfam_col = "bound_protein_pfam"
bound_prot_site_col = "bound_protein_site"

bound_prot_chainid_col = "bound_protein_chainid"
bound_interf_chainid_col = "bound_interf_chainid"

edia_col = "edia_score"
b_factor_col = "b_factor"

phi_col = "phi"
psi_col = "psi"
omega_col = "omega"
chi1_col = "chi1"
altchi1_col = "altchi1"
chi2_col = "chi2"
altchi2_col = "altchi2"
chi3_col = "chi3"
chi4_col = "chi4"
chi5_col = "chi5"

complete_col = "complete"
rama_col = "rama"
rotamer_col = "rotamer"
bb_resid_col = "bb_resid"
sc_resid_col = "sc_resid"
bb_seq_col = "bb_seq"
sc_seq_col = "sc_seq"
bb_len_col = "bb_length"
sc_len_col = "sc_length"

cluster_col = "cluster"
nn_dist_col = "nn_dist"
constr_dist_col = "constraint_distance"

nn_cutoff_col = "nn_cutoff"
constr_cutoff_col = "constraint_cutoff"

total_col = "total"
total_chain_col = "total_chain"
total_entry_col = "total_entry"
total_cf_col = "total_cf"

loop_col = "loop"
silh_col = "silhouette_score"
simi_col = "similarity_score"
total_complete_col = "total_complete"
cluster_count_col = "cluster_count"
total_cluster_pre_col = "total_cluster_pre"
total_noise_pre_col = "total_noise_pre"
total_pruned_nn_col = "total_pruned_nn"
total_pruned_constr_col = "total_pruned_constr"
total_classified_nn_col = "total_classified_nn"
total_classified_constr_col = "total_classified_constr"
total_cluster_post_col = "total_cluster_post"
total_noise_post_col = "total_noise_post"
select_col = "select"

rep_col = "representative"
common_col = "common"
entropy_col = "entropy"
occup_col = "occupancy"

mean_col = "mean"
max_col = "max"

vect_1_col = "vector_1"
vect_2_col = "vector_2"
atom_dist_col = "atom_dist"
hb_status_col = "hydrogen_bond"
hb_angle_1_col = "hb_angle_1"
hb_angle_2_col = "hb_angle_2"
wmhb_angle_col = "wmhb_angle"
outlier_col = "outlier"

bond_col = "bond"
angle_col = "angle"

dih_col = "dihedral"
rmsd_col = "rmsd"

index_col = "index"
p_col = "p_val"
correct_p_col = "corrected_p_val"

corr_col = "correlation"

a_col = "a_val"
b_col = "b_val"
c_col = "c_val"
d_col = "d_val"
risk_ratio_col = "risk_ratio"
low_ci_col = "lower_ci"
up_ci_col = "upper_ci"
sig_col = "significance"

bb_col_lst = [phi_col, psi_col, omega_col]

sc_col_lst = [
    chi1_col,
    chi2_col,
    altchi1_col,
    altchi2_col,
    chi3_col,
    chi4_col,
    chi5_col,
]

dih_col_lst = bb_col_lst + sc_col_lst

reformat_col_lst = dih_col_lst

dist_col_lst = [
    nn_dist_col,
    constr_dist_col,
    vect_1_col,
    vect_2_col,
    atom_dist_col,
    hb_angle_1_col,
    hb_angle_2_col,
    wmhb_angle_col,
    outlier_col,
]

sum_col_lst = [common_col, entropy_col, occup_col]

data_col_lst = reformat_col_lst + dist_col_lst + sum_col_lst

class_col_lst = [
    gene_class_col,
    nuc_class_col,
    mut_class_col,
    pharm_class_col,
    match_class_col,
    prot_class_col,
    interf_class_col,
    pocket_class_col,
]

path_col_lst = [
    core_path_col,
    rcsb_path_col,
    renum_path_col,
    rcsb_assembly_path_col,
    renum_assembly_path_col,
    sifts_path_col,
    edia_path_col,
    interf_path_col,
    pocket_path_col,
]

count_col_dict = {
    pdb_id_col: total_chain_col,
    pdb_code_col: total_entry_col,
    cf_col: total_cf_col,
}

order_col_lst = [
    id_col,
    pdb_id_col,
    pdb_code_col,
    modelid_col,
    chainid_col,
    sw1_col,
    sw2_col,
    y32_col,
    y71_col,
    cf_col,
    cluster_col,
    total_col,
    total_chain_col,
    total_entry_col,
    total_cf_col,
    rep_col,
    rama_col,
    rotamer_col,
    nn_dist_col,
    constr_dist_col,
    interf_id_col,
    interf_col,
    pocket_id_col,
    pocket_col,
    prot_col,
    swiss_id_col,
    uniprot_id_col,
    pfam_col,
    mut_status_col,
    mut_pos_col,
    gene_class_col,
    nuc_class_col,
    mut_class_col,
    pharm_class_col,
    match_class_col,
    prot_class_col,
    interf_class_col,
    pocket_class_col,
    loop_col,
    silh_col,
    simi_col,
    total_complete_col,
    cluster_count_col,
    total_cluster_pre_col,
    total_noise_pre_col,
    total_pruned_nn_col,
    total_pruned_constr_col,
    total_classified_nn_col,
    total_classified_constr_col,
    total_cluster_post_col,
    total_noise_post_col,
    select_col,
    nn_cutoff_col,
    constr_cutoff_col,
    space_col,
    len_col,
    range_col,
    date_col,
    pmid_col,
    index_col,
    p_col,
    correct_p_col,
    corr_col,
    risk_ratio_col,
    low_ci_col,
    up_ci_col,
    a_col,
    b_col,
    c_col,
    d_col,
    resid_col,
    resname_col,
    atomid_col,
    interf_area_col,
    cb_cont_col,
    atomid_cont_col,
    total_cb_cont_col,
    total_atomid_cont_col,
    interf_cont_col,
    cb_dist_col,
    iso_col,
    pocket_site_col,
    pocket_volume_col,
    pocket_score_col,
    pocket_status_col,
    pocket_type_col,
    pocket_cont_col,
    phi_col,
    psi_col,
    omega_col,
    chi1_col,
    chi2_col,
    altchi1_col,
    altchi2_col,
    chi3_col,
    chi4_col,
    chi5_col,
    edia_col,
    b_factor_col,
    vect_1_col,
    vect_2_col,
    atom_dist_col,
    hb_status_col,
    hb_angle_1_col,
    hb_angle_2_col,
    wmhb_angle_col,
    outlier_col,
    bond_col,
    angle_col,
    rmsd_col,
    common_col,
    entropy_col,
    occup_col,
    bound_lig_col,
    bound_prot_col,
    bound_lig_cont_col,
    bound_prot_cont_col,
    bound_prot_swiss_id_col,
    pharm_lig_site_col,
    pharm_lig_match_col,
    pharm_lig_smiles_col,
    bound_prot_pfam_col,
    bound_prot_site_col,
    bound_prot_chainid_col,
    bound_interf_chainid_col,
    bio_lig_col,
    ion_lig_col,
    pharm_lig_col,
    chem_lig_col,
    mod_lig_col,
    mem_lig_col,
    pocket_lig_col,
    method_col,
    resolution_col,
    r_factor_col,
    seq_col,
    bb_resid_col,
    bb_seq_col,
    bb_len_col,
    sc_resid_col,
    sc_seq_col,
    sc_len_col,
    complete_col,
    len_a_col,
    len_b_col,
    len_c_col,
    ang_a_col,
    ang_b_col,
    ang_g_col,
    core_path_col,
    rcsb_path_col,
    renum_path_col,
    rcsb_assembly_path_col,
    renum_assembly_path_col,
    sifts_path_col,
    edia_path_col,
    interf_path_col,
    pocket_path_col,
]

rename_col_dict = {
    id_col: "ID",
    pdb_id_col: "PDB ID",
    pdb_code_col: "PDB Code",
    modelid_col: "Model",
    chainid_col: "Chain",
    sw1_col: "SW1 Conformation",
    sw2_col: "SW2 Conformation",
    y32_col: "Y32 Position",
    y71_col: "Y71 Position",
    hb_status_col:"H-Bond Substate",
    gene_class_col: "RAS Isoform",
    mut_status_col: "Mutation Status",
    nuc_class_col: "Nucleotide State",
    prot_class_col: "Bound Protein",
    pocket_class_col: "Inhibitor Site",
    match_class_col: "Inhibitor Chemistry",
    interf_class_col: "Homodimer Status",
    bound_prot_col: "Bound Protein Name",
    bound_prot_swiss_id_col: "Bound Protein SwissProt ID",
    bound_prot_chainid_col: "Bound Protein Chain",
    bio_lig_col: "Nucleotide",
    ion_lig_col: "Ion",
    pharm_lig_col: "Inhibitor",
    chem_lig_col: "Chemical",
    mod_lig_col: "Modification",
    mem_lig_col: "Membrane",
    pocket_lig_col: "Pocket",
    space_col: "Space Group",
    method_col: "Experiment Method",
    resolution_col: "Resolution",
    r_factor_col: "R-Factor",
    date_col: "Deposit Date",
    pocket_score_col: "Druggability Scores",
    pocket_volume_col: "Pocket Volume"
}


def get_dist_col(x_resid, y_resid, x_atomid=None, y_atomid=None, ext=None):

    if ext is None:
        ext = atom_dist_col

    dist_col = str(x_resid)

    if str(x_atomid) != "None":
        dist_col += f"({x_atomid})"

    dist_col += ":"
    dist_col += str(y_resid)

    if str(y_atomid) != "None":
        dist_col += f"({y_atomid})"

    if ext is not None:
        dist_col += f"_{ext}"

    return dist_col



# -*- coding: utf-8 -*-
"""
  Copyright 2022 Mitchell Isaac Parker

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.

"""

import os
from tqdm import tqdm

from ..functions.path import (
    get_file_path,
    load_table,
    get_file_name,
    save_table,
    save_json,
    load_json,
    rascore_str,
    build_str,
)
from ..functions.file import (
    entry_table_file,
    interf_table_file,
    pocket_table_file,
    interf_json_file,
    pocket_json_file,
    dih_json_file,
)
from ..functions.col import path_col_lst


def prep_table(file_name, build_path=None):

    file_path = get_file_path(file_name, dir_path=build_path)

    df = load_table(file_path)

    if df is not None:
        df_col_lst = list(df.columns)
        for index in list(df.index.values):
            for col in [x for x in path_col_lst if x in df_col_lst]:
                dir_str = col.split("_path")[0]
                df.at[index, col] = get_file_path(
                    get_file_name(df.at[index, col]), dir_path=f"{build_path}/{dir_str}"
                )

        save_table(file_path, df)


def prep_json(file_name, build_path=None):

    file_path = get_file_path(file_name, dir_path=build_path)

    old_dict = load_json(file_path)

    if old_dict is not None:
        new_dict = dict()
        for old_key in list(old_dict.keys()):
            for col in path_col_lst:
                dir_str = col.split("_path")[0]
                if f"/{dir_str}/" in old_key:
                    break
            new_key = get_file_path(
                get_file_name(old_key), dir_path=f"{build_path}/{dir_str}"
            )
            new_dict[new_key] = old_dict[old_key]
            for sub_key in list(new_dict[new_key].keys()):
                for col in path_col_lst:
                    if col in list(new_dict[new_key][sub_key].keys()):
                        sub_dir_str = col.split("_path")[0]
                        new_dict[new_key][sub_key][col] = get_file_path(
                            get_file_name(new_dict[new_key][sub_key][col]),
                            dir_path=f"{build_path}/{sub_dir_str}",
                        )

        save_json(file_path, new_dict)


def prep_rascore(build_path=None):

    if build_path is None:
        build_path = f"{os.getcwd()}/{rascore_str}_{build_str}"

    table_file_lst = [entry_table_file, interf_table_file, pocket_table_file]

    json_file_lst = [interf_json_file, pocket_json_file, dih_json_file]

    for table_file in tqdm(
        table_file_lst,
        desc="Preparing Rascore database tables",
        position=0,
        leave=True,
    ):
        prep_table(table_file, build_path=build_path)

    for json_file in tqdm(
        json_file_lst,
        desc="Preparing Rascore database jsons",
        position=0,
        leave=True,
    ):
        prep_json(json_file, build_path=build_path)



# -*- coding: utf-8 -*-
"""
  Copyright 2022 Mitchell Isaac Parker

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.

"""

import pandas as pd
import numpy as np
from tqdm import tqdm

from ..functions.col import (
    core_path_col,
    pocket_volume_col,
    pocket_score_col,
    pocket_cont_col,
    pocket_col,
    pocket_id_col,
    pdb_id_col,
)
from ..functions.table import get_df_at_index, fix_col
from ..functions.lst import str_to_lst, calc_jaccard, calc_simpson
from ..functions.path import save_table


def get_index_pocket(
    df,
    index,
    pocket_dict,
    min_volume=None,
    min_score=None,
    search_cont_lst=None,
    search_max_dist=0.7,
    coord_path_col=None,
    use_simpson=False,
):

    if coord_path_col is None:
        coord_path_col = core_path_col

    if search_cont_lst is not None:
        if not any(isinstance(x, list) for x in search_cont_lst):
            search_cont_lst = [search_cont_lst]
            for i, search_cont in enumerate(search_cont_lst):
                search_cont_lst[i] = [str(x) for x in search_cont]

    index_df = get_df_at_index(df, index)

    coord_path = index_df.at[index, coord_path_col]
    coord_path = coord_path.replace(".cif", ".pdb")

    pocket_df = pd.DataFrame()

    if coord_path in list(pocket_dict.keys()):
        for pocket in list(pocket_dict[coord_path].keys()):

            add_pocket = True

            if min_volume is not None:
                if (
                    float(pocket_dict[coord_path][pocket][pocket_volume_col])
                    < min_volume
                ):
                    add_pocket = False

            if min_score is not None:
                if float(pocket_dict[coord_path][pocket][pocket_score_col]) < min_score:
                    add_pocket = False

            if search_cont_lst is not None:

                pocket_cont = str_to_lst(
                    pocket_dict[coord_path][pocket][pocket_cont_col]
                )

                pocket_dist_lst = list()
                for search_cont in search_cont_lst:
                    if use_simpson:
                        pocket_dist = calc_simpson(
                            pocket_cont,
                            search_cont,
                            return_dist=True,
                        )

                    else:
                        pocket_dist = calc_jaccard(
                            pocket_cont,
                            search_cont,
                            return_dist=True,
                        )
                    pocket_dist_lst.append(pocket_dist)

                if np.mean(pocket_dist_lst) >= search_max_dist:
                    add_pocket = False

            if add_pocket:
                temp_df = index_df.copy(deep=True)

                for col in list(pocket_dict[coord_path][pocket].keys()):
                    temp_df.at[index, col] = pocket_dict[coord_path][pocket][col]

                temp_df.at[index, pocket_col] = pocket

                pocket_df = pd.concat([pocket_df, temp_df], sort=False)

    return pocket_df


def build_pocket_table(
    df,
    pocket_dict,
    pocket_table_path=None,
    min_volume=None,
    min_score=None,
    search_cont_lst=None,
    search_max_dist=0.7,
    use_simpson=False,
    coord_path_col=None,
):

    df = df.reset_index(drop=True)

    if coord_path_col is None:
        coord_path_col = core_path_col

    pocket_df = pd.DataFrame()

    for index in tqdm(
        list(df.index.values), desc="Building pocket table", position=0, leave=True
    ):

        pocket_df = pd.concat(
            [
                pocket_df,
                get_index_pocket(
                    df,
                    index,
                    pocket_dict,
                    min_volume=min_volume,
                    min_score=min_score,
                    search_cont_lst=search_cont_lst,
                    search_max_dist=search_max_dist,
                    use_simpson=use_simpson,
                    coord_path_col=coord_path_col,
                ),
            ],
            sort=False,
        )

    if len(pocket_df) > 0:
        pocket_df = pocket_df.reset_index(drop=True)

        pocket_df[pocket_col] = pocket_df[pocket_col].map(str)
        pocket_df = fix_col(pocket_df, pocket_col)

        pocket_df[pocket_id_col] = pocket_df[pdb_id_col] + pocket_df[pocket_col]

    if pocket_table_path is not None:
        save_table(pocket_table_path, pocket_df)
    else:
        return pocket_df

    print("Built pocket table!")



# -*- coding: utf-8 -*-
"""
  Copyright 2022 Mitchell Isaac Parker

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.

"""

import streamlit as st
from PIL import Image

from ..constants.conf import (
    loop_color_dict,
)
from ..functions.col import (
    rename_col_dict,
    nuc_class_col,
    pocket_class_col,
    prot_class_col,
)
from ..constants.nuc import nuc_color_dict
from ..constants.pharm import pharm_color_dict, sp2_name, sp12_name
from ..constants.prot import (
    prot_color_dict,
    effector_name,
    gef_cdc_name,
    gef_rem_name,
    gap_name,
    binder_name,
    nano_name,
)
from ..functions.gui import (
    load_st_table,
    get_html_text,
    write_st_end,
    get_neighbor_path,
)
from ..functions.path import get_file_path, pages_str, data_str


def overview_page():

    df = load_st_table(__file__)

    st.markdown("## Database Overview")

    st.markdown("---")

    st.sidebar.markdown(
        "**Note.** See our [bioRxiv](https://www.biorxiv.org/content/10.1101/2022.02.02.478568v2) paper for additional details."
    )

    text_col, leg_1_col, leg_2_col = st.columns((2, 0.75, 0.75))

    text_col.markdown("#### Introduction to the Rascore Database")
    text_col.markdown(
        f"""
    - For many human cancers and tumor-associated diseases, mutations in the RAS isoforms (KRAS, NRAS, and HRAS) 
    are the most common oncogenic alterations, making these proteins high-priority therapeutic targets. 
    - Effectively targeting the RAS isoforms requires an exact understanding of 
    their active, inactive, and druggable conformations.  
    - In consequence, we analyzed over 700 available human KRAS, NRAS, and HRAS structures in the 
    Protein Data Bank (PDB) to create a comprehensive map of the RAS conformational landscape. 
      - First, we annotated the molecular contents of each RAS structure, including their mutation status, 
    nucleotide state, and bound protein or inhibitor site (*see right for key terms*). 
      - Second, we conformationally classified all available RAS structures based on the configurations 
      of their catalytic switch 1 (SW1) and switch 2 (SW2) loops, identifying three SW1 and nine SW2 conformations.
    - The *Rascore* database presents a continually updated dataset of RAS structures in the PDB that are
    conformationally classified and annotated for their molecular contents (*now {len(df)} structures*).
    """
    )

    loop_term_dict = {"SW1": "Switch 1", "SW2": "Switch 2"}

    nuc_term_dict = {
        "0P": "Nucleotide-free",
        "2P": "GDP-bound",
        "3P": "GTP or GTP analog-bound",
    }

    prot_term_dict = {
        effector_name: "RAS-binding domain (RBD) or RAS-associating domain (RA) protein",
        gef_cdc_name: "Guanine exchange factor, catalaytic domain",
        gef_rem_name: "Guanine exchange factor, allosteric domain",
        gap_name: "GTPase activating protein",
        binder_name: "Designed protein inhibitor",
        nano_name: "Synthetic membrane",
    }

    pharm_term_dict = {
        sp12_name: "SW1/SW2 pocket",
        sp2_name: "SW2 pocket",
    }

    loop_str = "**Loop Names**"
    nuc_str = f"**{rename_col_dict[nuc_class_col]}**"
    pharm_str = f"**{rename_col_dict[pocket_class_col]}**"
    prot_str = f"**{rename_col_dict[prot_class_col]}**"

    leg_info_dict = {
        loop_str: loop_term_dict,
        nuc_str: nuc_term_dict,
        pharm_str: pharm_term_dict,
        prot_str: prot_term_dict,
    }
    leg_color_dict = {
        loop_str: loop_color_dict,
        nuc_str: nuc_color_dict,
        pharm_str: pharm_color_dict,
        prot_str: prot_color_dict,
    }
    leg_col_dict = {
        loop_str: leg_1_col,
        nuc_str: leg_1_col,
        pharm_str: leg_1_col,
        prot_str: leg_2_col,
    }

    for info in list(leg_info_dict.keys()):

        col = leg_col_dict[info]

        col.markdown(info)

        for term, desc in leg_info_dict[info].items():
            col.markdown(
                get_html_text(
                    {
                        term: leg_color_dict[info][term],
                        " - ": "#31333F",
                        desc: "#31333F",
                    },
                    font_size="small",
                ),
                unsafe_allow_html=True,
            )

    st.markdown("---")

    st.markdown("#### Our Conformational Classification Algorithm")

    st.markdown(
        """
    - First, we broadly classified RAS structures based on the spatial positions of residue Y32 relative to the 
    active site (“Y32in” or “Y32out”) in SW1 and residue Y71 relative to the hydrophobic 
    core (“Y71in” or “Y71out”) in SW2. 
    - Second, we separately clustered the configurations of SW1 (residues 25-40) and SW2 (residues 56-76)
    based on their backbone dihedral angle values: φ (phi), ψ (psi), and ω (omega).
    - In clustering RAS SW1 and SW2 loops, we used the Density-Based Spatial Clustering of Applications 
    with Noise (DBSCAN) algorithm with a distance metric that locates the maximum backbone dihedral angle 
    difference upon pairwise comparison of loop residues (previously implemented by our group for other proteins, 
    such as [kinases](http://dunbrack.fccc.edu/kincore/home) and [antibodies](http://dunbrack2.fccc.edu/PyIgClassify/)).
      - DBSCAN clusters points with sufficient numbers of near neighbors and classifies the remainder as outliers. 
      - We first separated RAS structures by nucleotide state (0P, 2P, and 3P) and spatial class 
      (Y32in/out for SW1 and Y71in/out for SW2) and, within each group, subsequently clustered the 
      conformations of completely modeled SW1 and SW2 loops with well-defined electron density. 
      - We then assigned a small number of poorly or incompletely modeled loops to the clusters obtained from 
        DBSCAN by using a nearest neighbors (NN) approach that we developed.
      - In the Rascore database, we use our NN approach to conformationally classify user uploaded 
        structures and additional RAS structures deposited to the PDB.
    """
    )

    st.markdown("---")

    st.markdown("#### Summary of the Current RAS Conformational Landscape")

    st.markdown("""
    - For clarity and brevity in our RAS conformational classification, 
    we named each SW1 and SW2 conformation by its *spatial class*, *nucleotide state*, 
    and a *conformational label* (written in all-capital letters). 
    - The SW1 conformations are labeled Y32in.3P-ON (GTP-bound state 2), 
    Y32out.2P-OFF (GDP-bound), and Y32out.0P-GEF (nucleotide-free) (*panel* **A**).
    - There was no structurally uniform cluster within Y32out.3P structures 
    that could be called the GTP-bound state 1. 
    - The only nucleotide-free SW2 conformation was Y32out.0P-GEF (*panel* **B**).
    - The GTP-bound SW2 conformations included Y71in.3P-R (R state) and Y71out.3P-T (T state), 
    and two previously unclassified druggable conformations associated with inhibitors at the 
    SP12 site, which we named Y71in.3P-SP12-A and Y71.3P-SP12-B (*panel* **C**). 
    - The four GDP-bound SW2 conformations are all named for their predominant binding partners, 
    which consist of SP2 and SP12 inhibitors and protein binders: Y71out.2P-SP2-A, Y71out.2P-SP2-B, 
    Y71in.2P-SP12, and Y71out.2P-BINDER (*panel* **D**).
    - In *panels* **E** and **F**, our original SW1 and SW2 conformational clustering (with NN assignments added) 
    are displayed as Ramachandran maps per residue of each cluster.
    """)

    img = Image.open(
        get_file_path(
            "rascore_figure.png",
            dir_path=get_neighbor_path(__file__, pages_str, data_str),
        )
    )

    st.image(img, output_format="PNG")


    write_st_end()



from src.download.modules import *

from src.renum.shared.handling_chain_numbering_clashes import handling_chain_numbering_clashes
from src.renum.shared.renumbered_count_in_chains import renumbered_count_in_chains
from src.download.downloadwithThreadPool import download_with_pool, url_formation_for_pool
from src.download import compressor


def try_MMCIF2Dict(default_input_path_to_mmCIF, mmCIF_name):
    mmcif_dict = 0
    for _ in range(3):
        try:
            mmcif_dict = Bio.PDB.MMCIF2Dict.MMCIF2Dict(gzip.open(Path(str(default_input_path_to_mmCIF) + "/" + mmCIF_name), 'rt'))
            break
        except EOFError:
            os.remove(Path(str(default_input_path_to_mmCIF) + "/" + mmCIF_name))
            if "assembly" in mmCIF_name:
                download_with_pool(url_formation_for_pool("mmCIF_assembly", [mmCIF_name])[0])
            else:
                download_with_pool(url_formation_for_pool("mmCIF", [mmCIF_name])[0])
        except ValueError:
            os.remove(Path(str(default_input_path_to_mmCIF) + "/" + mmCIF_name))
            if "assembly" in mmCIF_name:
                download_with_pool(url_formation_for_pool("mmCIF_assembly", [mmCIF_name])[0])
            else:
                download_with_pool(url_formation_for_pool("mmCIF", [mmCIF_name])[0])
        except OSError:
            if "assembly" in mmCIF_name:
                download_with_pool(url_formation_for_pool("mmCIF_assembly", [mmCIF_name])[0])
            else:
                download_with_pool(url_formation_for_pool("mmCIF", [mmCIF_name])[0])
    return mmcif_dict


def try_SIFTS_tree_parser(default_input_path_to_SIFTS, SIFTS_name):
    product_tree_SIFTS = 0
    for _ in range(3):
        try:
            handle_SIFTS = gzip.open(Path(str(default_input_path_to_SIFTS) + "/" + SIFTS_name), 'rt')
            product_tree_SIFTS = SIFTS_tree_parser(handle_SIFTS)
            break
        except EOFError:
            os.remove(Path(str(default_input_path_to_SIFTS) + "/" + SIFTS_name))
            download_with_pool(url_formation_for_pool("SIFTS", [SIFTS_name])[0])
        except ValueError:
            os.remove(Path(str(default_input_path_to_SIFTS) + "/" + SIFTS_name))
            download_with_pool(url_formation_for_pool("SIFTS", [SIFTS_name])[0])
        except OSError:
            download_with_pool(url_formation_for_pool("SIFTS", [SIFTS_name])[0])
        except Exception:
            download_with_pool(url_formation_for_pool("SIFTS", [SIFTS_name])[0])
    return product_tree_SIFTS


def output_with_this_name_ending(name_ending, path, mmcif_dict, mmCIF_name, gzip_mode, current_directory):
    mmCIF_name = mmCIF_name[:mmCIF_name.rfind(".cif.gz")]
    os.chdir(path)
    io = MMCIFIO()
    io.set_dict(mmcif_dict)
    io.save(mmCIF_name + name_ending)
    if gzip_mode == "on":
        compressor.compress_output_files(mmCIF_name + name_ending, gzip_mode)
        os.remove(mmCIF_name + name_ending)
    os.chdir(current_directory)


def copy_file(inpath, file_name, outpath, postfix, gzip_mode):
    mmCIF_name = file_name[:file_name.rfind(".cif.gz")]
    absolute_path_in = inpath + "/" + file_name
    absolute_path_out = outpath + "/" + mmCIF_name + postfix
    if gzip_mode == "off":
        with gzip.open(absolute_path_in, 'rb') as f_in:
            with open(absolute_path_out[:-3], 'wb') as f_out:
                shutil.copyfileobj(f_in, f_out)
    else:
        shutil.copyfile(absolute_path_in, absolute_path_out)


def if_no_SIFTS_data_log(mmCIF_name, mmcif_dict, log_message):
    strand_id_set = set()
    try:
        pull_chains_for_chains_count = mmcif_dict["_pdbx_poly_seq_scheme.pdb_strand_id"]
    except KeyError:
        try:
            pull_chains_for_chains_count = mmcif_dict["_pdbe_orig_poly_seq_scheme.pdb_strand_id"]
        except KeyError:
            pull_chains_for_chains_count = mmcif_dict["_atom_site.auth_asym_id"]

    for strand in pull_chains_for_chains_count:
        strand_id_set.add(strand)
    strand_id_set = list(strand_id_set)
    strand_id_set.sort()
    for strand in strand_id_set:
        count_elements_in_strand = 0
        for chain_id in pull_chains_for_chains_count:
            if chain_id == strand:
                count_elements_in_strand += 1
        log_message.append([mmCIF_name[:4], strand, "-", "-", "-", "-", count_elements_in_strand, "0", "0"])
    return log_message


def renum_struct_ref_seq_pdbx_auth_seq_align(mmcif_dict):
    try:
        _struct_ref_seq_pdbx_strand_id = mmcif_dict["_struct_ref_seq.pdbx_strand_id"]

        _struct_ref_seq_pdbx_seq_align_beg_ins_code = mmcif_dict["_struct_ref_seq.pdbx_seq_align_beg_ins_code"]
        _struct_ref_seq_pdbx_auth_seq_align_beg = mmcif_dict["_struct_ref_seq.pdbx_auth_seq_align_beg"]
        _struct_ref_seq_db_align_beg = mmcif_dict["_struct_ref_seq.db_align_beg"]
        mmcif_dict["_struct_ref_seq.pdbx_auth_seq_align_beg"] = mmcif_dict["_struct_ref_seq.db_align_beg"]

        _struct_ref_seq_pdbx_seq_align_end_ins_code = mmcif_dict["_struct_ref_seq.pdbx_seq_align_end_ins_code"]
        _struct_ref_seq_pdbx_auth_seq_align_end = mmcif_dict["_struct_ref_seq.pdbx_auth_seq_align_end"]
        _struct_ref_seq_db_align_end = mmcif_dict["_struct_ref_seq.db_align_end"]
        mmcif_dict["_struct_ref_seq.pdbx_auth_seq_align_end"] = mmcif_dict["_struct_ref_seq.db_align_end"]

        if type(_struct_ref_seq_pdbx_seq_align_beg_ins_code) == str:
            if "." in _struct_ref_seq_pdbx_seq_align_beg_ins_code:
                mmcif_dict["_struct_ref_seq.pdbx_seq_align_beg_ins_code"] = "."
            else:
                mmcif_dict["_struct_ref_seq.pdbx_seq_align_beg_ins_code"] = "?"
        if type(_struct_ref_seq_pdbx_seq_align_end_ins_code) == str:
            if "." in _struct_ref_seq_pdbx_seq_align_end_ins_code:
                mmcif_dict["_struct_ref_seq.pdbx_seq_align_end_ins_code"] = "."
            else:
                mmcif_dict["_struct_ref_seq.pdbx_seq_align_end_ins_code"] = "?"

        PDB_ins_code_list = list()
        if type(_struct_ref_seq_pdbx_seq_align_beg_ins_code) != str:
            if "." in _struct_ref_seq_pdbx_seq_align_beg_ins_code:
                for _ in range(len(_struct_ref_seq_pdbx_seq_align_beg_ins_code)):
                    PDB_ins_code_list.append(".")
            else:
                for _ in range(len(_struct_ref_seq_pdbx_seq_align_beg_ins_code)):
                    PDB_ins_code_list.append("?")
            mmcif_dict["_struct_ref_seq.pdbx_seq_align_beg_ins_code"] = PDB_ins_code_list
            mmcif_dict["_struct_ref_seq.pdbx_seq_align_end_ins_code"] = PDB_ins_code_list

    except KeyError:
        pass


def poly_nonpoly_renum(mmcif_dict, df_PDBe_PDB_UniProt, chains_to_change, default_mmCIF_num):
    try:
        _pdbx_poly_seq_scheme_seq_id = mmcif_dict["_pdbx_poly_seq_scheme.seq_id"]
        _pdbx_poly_seq_scheme_asym_id = mmcif_dict["_pdbx_poly_seq_scheme.asym_id"]
        _pdbx_poly_seq_scheme_mon_id = mmcif_dict["_pdbx_poly_seq_scheme.mon_id"]

        _pdbx_poly_seq_scheme_pdb_seq_num = mmcif_dict["_pdbx_poly_seq_scheme.pdb_seq_num"]
        _pdbx_poly_seq_scheme_auth_seq_num = mmcif_dict["_pdbx_poly_seq_scheme.auth_seq_num"]
        _pdbx_poly_seq_scheme_pdb_mon_id = mmcif_dict["_pdbx_poly_seq_scheme.pdb_mon_id"]
        _pdbx_poly_seq_scheme_auth_mon_id = mmcif_dict["_pdbx_poly_seq_scheme.auth_mon_id"]
        _pdbx_poly_seq_scheme_pdb_strand_id = mmcif_dict["_pdbx_poly_seq_scheme.pdb_strand_id"]
        _pdbx_poly_seq_scheme_pdb_ins_code = mmcif_dict["_pdbx_poly_seq_scheme.pdb_ins_code"]
    except KeyError:
        try:
            _pdbx_poly_seq_scheme_seq_id = mmcif_dict["_pdbe_orig_poly_seq_scheme.seq_id"]
            _pdbx_poly_seq_scheme_asym_id = mmcif_dict["_pdbe_orig_poly_seq_scheme.asym_id"]
            _pdbx_poly_seq_scheme_mon_id = mmcif_dict["_pdbe_orig_poly_seq_scheme.mon_id"]

            _pdbx_poly_seq_scheme_pdb_seq_num = mmcif_dict["_pdbe_orig_poly_seq_scheme.pdb_seq_num"]
            _pdbx_poly_seq_scheme_auth_seq_num = mmcif_dict["_pdbe_orig_poly_seq_scheme.auth_seq_num"]
            _pdbx_poly_seq_scheme_pdb_mon_id = mmcif_dict["_pdbe_orig_poly_seq_scheme.pdb_mon_id"]
            _pdbx_poly_seq_scheme_auth_mon_id = mmcif_dict["_pdbe_orig_poly_seq_scheme.auth_mon_id"]
            _pdbx_poly_seq_scheme_pdb_strand_id = mmcif_dict["_pdbe_orig_poly_seq_scheme.pdb_strand_id"]
            _pdbx_poly_seq_scheme_pdb_ins_code = mmcif_dict["_pdbe_orig_poly_seq_scheme.pdb_ins_code"]

        except KeyError:
            # continue
            return 0

    if type(_pdbx_poly_seq_scheme_pdb_strand_id) == str:
        _pdbx_poly_seq_scheme_pdb_seq_num = [_pdbx_poly_seq_scheme_pdb_seq_num]
        _pdbx_poly_seq_scheme_auth_seq_num = [_pdbx_poly_seq_scheme_auth_seq_num]
        _pdbx_poly_seq_scheme_pdb_mon_id = [_pdbx_poly_seq_scheme_pdb_mon_id]
        _pdbx_poly_seq_scheme_auth_mon_id = [_pdbx_poly_seq_scheme_auth_mon_id]
        _pdbx_poly_seq_scheme_pdb_strand_id = [_pdbx_poly_seq_scheme_pdb_strand_id]
        _pdbx_poly_seq_scheme_pdb_ins_code = [_pdbx_poly_seq_scheme_pdb_ins_code]

    mmCIF_pdbx_poly_seq_scheme_label = list(zip(_pdbx_poly_seq_scheme_seq_id,
                                                _pdbx_poly_seq_scheme_mon_id,
                                                _pdbx_poly_seq_scheme_asym_id))
    mmCIF_pdbx_poly_seq_scheme_pdb = list(zip(_pdbx_poly_seq_scheme_pdb_seq_num,
                                              _pdbx_poly_seq_scheme_pdb_mon_id,
                                              _pdbx_poly_seq_scheme_pdb_strand_id))
    mmCIF_pdbx_poly_seq_scheme_auth = list(zip(_pdbx_poly_seq_scheme_auth_seq_num,
                                               _pdbx_poly_seq_scheme_auth_mon_id,
                                               _pdbx_poly_seq_scheme_pdb_strand_id))

    df_mmCIF_pdbx_poly_seq_scheme = pd.DataFrame(zip(mmCIF_pdbx_poly_seq_scheme_label,
                                                     mmCIF_pdbx_poly_seq_scheme_pdb,
                                                     mmCIF_pdbx_poly_seq_scheme_auth,
                                                     _pdbx_poly_seq_scheme_pdb_ins_code))

    df_mmCIF_pdbx_poly_seq_scheme = df_mmCIF_pdbx_poly_seq_scheme.rename(
        columns={0: "_pdbx_poly_seq_scheme_label", 1: "pdbx_poly_seq_scheme_pdb",
                 2: "pdbx_poly_seq_scheme_auth", 3: "pdbx_poly_seq_scheme_pdb_ins_code"})

    df_pdbx_poly_seq_scheme_pdb_final = df_mmCIF_pdbx_poly_seq_scheme.merge(
        df_PDBe_PDB_UniProt, left_on="_pdbx_poly_seq_scheme_label", right_on="PDBe", how='left')
    df_pdbx_poly_seq_scheme_pdb_final["PDBe_num_and_chain"] = df_pdbx_poly_seq_scheme_pdb_final[
        "_pdbx_poly_seq_scheme_label"].apply(lambda x: (x[0], x[2]))

    df_pdbx_poly_seq_scheme_pdb_final["PDB_num_and_chain"] = np.where(
        df_pdbx_poly_seq_scheme_pdb_final["pdbx_poly_seq_scheme_pdb_ins_code"].apply(lambda x: x == "."),
        df_pdbx_poly_seq_scheme_pdb_final["pdbx_poly_seq_scheme_pdb"].apply(lambda x: (x[0], x[2])),
        df_pdbx_poly_seq_scheme_pdb_final["pdbx_poly_seq_scheme_pdb"].apply(lambda x: x[0]) +
        df_pdbx_poly_seq_scheme_pdb_final["pdbx_poly_seq_scheme_pdb_ins_code"].apply(lambda x: x) + "," +
        df_pdbx_poly_seq_scheme_pdb_final["pdbx_poly_seq_scheme_pdb"].apply(lambda x: x[2]))
    df_pdbx_poly_seq_scheme_pdb_final["PDB_num_and_chain"] = df_pdbx_poly_seq_scheme_pdb_final["PDB_num_and_chain"].apply(
        lambda x: tuple(x.split(",")) if type(x) == str else x)

    df_pdbx_poly_seq_scheme_pdb_final["Uni_or_50k"] = np.where(
        df_pdbx_poly_seq_scheme_pdb_final["PDB_num_and_chain"].apply(lambda x: x[1] in chains_to_change),
        df_pdbx_poly_seq_scheme_pdb_final["UniProt_50k"].apply(lambda x: x),
        df_pdbx_poly_seq_scheme_pdb_final["PDB_num_and_chain"].apply(lambda x: x[0].strip(re.sub('[0-9\-\?\.]+', '', x[0]))))

    try:
        mmcif_dict["_pdbx_poly_seq_scheme.pdb_seq_num"]  # check if key exists
        mmcif_dict["_pdbx_poly_seq_scheme.pdb_seq_num"] = list(df_pdbx_poly_seq_scheme_pdb_final["Uni_or_50k"].values)
        mmcif_dict["_pdbx_poly_seq_scheme.auth_seq_num"] = _pdbx_poly_seq_scheme_pdb_seq_num
    except KeyError:
        mmcif_dict["_pdbe_orig_poly_seq_scheme.pdb_seq_num"] = list(df_pdbx_poly_seq_scheme_pdb_final["Uni_or_50k"].values)
        mmcif_dict["_pdbe_orig_poly_seq_scheme.auth_seq_num"] = _pdbx_poly_seq_scheme_pdb_seq_num

    nonpoly_present = False

    try:
        _pdbx_nonpoly_scheme_pdb_seq_num = mmcif_dict["_pdbx_nonpoly_scheme.pdb_seq_num"]
        _pdbx_nonpoly_scheme_auth_seq_num = mmcif_dict["_pdbx_nonpoly_scheme.auth_seq_num"]
        _pdbx_nonpoly_scheme_pdb_mon_id = mmcif_dict["_pdbx_nonpoly_scheme.pdb_mon_id"]
        _pdbx_nonpoly_scheme_auth_mon_id = mmcif_dict["_pdbx_nonpoly_scheme.auth_mon_id"]
        _pdbx_nonpoly_scheme_pdb_strand_id = mmcif_dict["_pdbx_nonpoly_scheme.pdb_strand_id"]
        _pdbx_nonpoly_scheme_asym_id = mmcif_dict["_pdbx_nonpoly_scheme.asym_id"]
        dots_for_label = ["." for _ in range(len(_pdbx_nonpoly_scheme_asym_id)) if type(_pdbx_nonpoly_scheme_asym_id) == list]
        nonpoly_present = True
    except KeyError:
        try:
            _pdbx_nonpoly_scheme_pdb_seq_num = mmcif_dict["_pdbe_orig_nonpoly_scheme.pdb_seq_num"]
            _pdbx_nonpoly_scheme_auth_seq_num = mmcif_dict["_pdbe_orig_nonpoly_scheme.auth_seq_num"]
            _pdbx_nonpoly_scheme_pdb_mon_id = mmcif_dict["_pdbe_orig_nonpoly_scheme.pdb_mon_id"]
            _pdbx_nonpoly_scheme_auth_mon_id = mmcif_dict["_pdbe_orig_nonpoly_scheme.auth_mon_id"]
            _pdbx_nonpoly_scheme_pdb_strand_id = mmcif_dict["_pdbe_orig_nonpoly_scheme.pdb_strand_id"]
            _pdbx_nonpoly_scheme_asym_id = mmcif_dict["_pdbe_orig_nonpoly_scheme.asym_id"]
            dots_for_label = ["." for _ in range(len(_pdbx_nonpoly_scheme_asym_id)) if type(_pdbx_nonpoly_scheme_asym_id) == list]
            nonpoly_present = True
        except KeyError:
            pass

    if nonpoly_present:
        if type(_pdbx_nonpoly_scheme_pdb_strand_id) == str:
            _pdbx_nonpoly_scheme_pdb_seq_num = [_pdbx_nonpoly_scheme_pdb_seq_num]
            _pdbx_nonpoly_scheme_auth_seq_num = [_pdbx_nonpoly_scheme_auth_seq_num]
            _pdbx_nonpoly_scheme_pdb_mon_id = [_pdbx_nonpoly_scheme_pdb_mon_id]
            _pdbx_nonpoly_scheme_auth_mon_id = [_pdbx_nonpoly_scheme_auth_mon_id]
            _pdbx_nonpoly_scheme_pdb_strand_id = [_pdbx_nonpoly_scheme_pdb_strand_id]
            _pdbx_nonpoly_scheme_asym_id = [_pdbx_nonpoly_scheme_asym_id]
            dots_for_label = ["."]

        mmCIF_pdbx_nonpoly_scheme_pdb = list(zip(_pdbx_nonpoly_scheme_pdb_seq_num,
                                                 _pdbx_nonpoly_scheme_pdb_mon_id,
                                                 _pdbx_nonpoly_scheme_pdb_strand_id))
        mmCIF_pdbx_nonpoly_scheme_auth = list(zip(_pdbx_nonpoly_scheme_auth_seq_num,
                                                  _pdbx_nonpoly_scheme_auth_mon_id,
                                                  _pdbx_nonpoly_scheme_pdb_strand_id))
        mmCIF_pdbx_nonpoly_scheme_label = list(zip(dots_for_label,
                                                   _pdbx_nonpoly_scheme_pdb_mon_id,
                                                   _pdbx_nonpoly_scheme_asym_id))

        df_mmCIF_pdbx_nonpoly_scheme = pd.DataFrame(zip(mmCIF_pdbx_nonpoly_scheme_pdb,
                                                        mmCIF_pdbx_nonpoly_scheme_auth,
                                                        mmCIF_pdbx_nonpoly_scheme_label))
        df_mmCIF_pdbx_nonpoly_scheme = df_mmCIF_pdbx_nonpoly_scheme.rename(columns={0: "pdbx_nonpoly_scheme_pdb",
                                                                                    1: "pdbx_nonpoly_scheme_auth",
                                                                                    2: "pdbx_nonpoly_scheme_label"})

        df_mmCIF_pdbx_nonpoly_scheme["PDB"] = df_mmCIF_pdbx_nonpoly_scheme["pdbx_nonpoly_scheme_pdb"]
        df_mmCIF_pdbx_nonpoly_scheme["PDB_num_and_chain"] = df_mmCIF_pdbx_nonpoly_scheme["pdbx_nonpoly_scheme_pdb"].apply(lambda x: (x[0], x[2]))
        df_mmCIF_pdbx_nonpoly_scheme["Uni_or_50k"] = df_mmCIF_pdbx_nonpoly_scheme["pdbx_nonpoly_scheme_pdb"].apply(
            lambda x: str(int(x[0]) + default_mmCIF_num + 10000) if x[2] in chains_to_change else x[0])

        try:
            mmcif_dict["_pdbx_nonpoly_scheme.pdb_seq_num"]  # check if key exists
            mmcif_dict["_pdbx_nonpoly_scheme.pdb_seq_num"] = list(df_mmCIF_pdbx_nonpoly_scheme["Uni_or_50k"].values)
            mmcif_dict["_pdbx_nonpoly_scheme.auth_seq_num"] = _pdbx_nonpoly_scheme_pdb_seq_num
        except KeyError:
            try:
                mmcif_dict["_pdbe_orig_nonpoly_scheme.pdb_seq_num"] = list(df_mmCIF_pdbx_nonpoly_scheme["Uni_or_50k"].values)
                mmcif_dict["_pdbe_orig_nonpoly_scheme.auth_seq_num"] = _pdbx_nonpoly_scheme_pdb_seq_num
            except KeyError:
                pass

        poly_nonpoly_append = df_pdbx_poly_seq_scheme_pdb_final.append(df_mmCIF_pdbx_nonpoly_scheme)
        poly_nonpoly_append = poly_nonpoly_append[["PDBe", "PDB", "UniProt", "PDBe_num_and_chain", "PDB_num_and_chain", "AccessionID", "Uni_or_50k"]]
    else:
        poly_nonpoly_append = df_pdbx_poly_seq_scheme_pdb_final[
            ["PDBe", "PDB", "UniProt", "PDBe_num_and_chain", "PDB_num_and_chain", "AccessionID", "Uni_or_50k"]]

    return poly_nonpoly_append


def renumber_tables(formed_columns, mmcif_dict, poly_nonpoly_atom_site, chains_to_change, default_mmCIF_num):
    dot_or_question_tuple = (".", "?")
    for n in formed_columns:
        auth_comp_id = 0
        auth_seq_id = n[0]
        auth_asym_id = n[1]
        try:
            PDB_ins_code = n[2]
            if "ins_code" not in PDB_ins_code:
                auth_comp_id = PDB_ins_code
                PDB_ins_code = 0
        except IndexError:
            PDB_ins_code = 0
        try:
            if auth_comp_id == 0:
                auth_comp_id = n[3]
        except IndexError:
            auth_comp_id = 0

        if "_pdbx_branch_scheme" in auth_seq_id:
            auth_seq_id = "_pdbx_branch_scheme.pdb_seq_num"
            auth_asym_id = "_pdbx_branch_scheme.pdb_asym_id"

        PDB_ins_code_list = list()
        # auth_comp_id_list = mmcif_dict[auth_comp_id] #for debug only
        auth_seq_id_list = mmcif_dict[auth_seq_id]
        auth_asym_id_list = mmcif_dict[auth_asym_id]

        if PDB_ins_code == 0:
            for _ in range(len(auth_seq_id_list)):
                PDB_ins_code_list.append("?")
        else:
            PDB_ins_code_list = mmcif_dict[PDB_ins_code]

        if type(auth_asym_id_list) == str:
            # auth_comp_id_list = [auth_comp_id_list] for debug only
            auth_seq_id_list = [auth_seq_id_list]
            auth_asym_id_list = [auth_asym_id_list]

            if PDB_ins_code == 0:
                PDB_ins_code_list = ["?"]
            else:
                PDB_ins_code_list = [PDB_ins_code]

        if PDB_ins_code != 0:
            dot_to_question = list()
            for ins_code in mmcif_dict[PDB_ins_code]:
                if ins_code == ".":
                    dot_to_question.append("?")
                else:
                    dot_to_question.append(ins_code)
            PDB_ins_code_list = dot_to_question

        auth_seq_id_list_zip = list(zip(auth_seq_id_list, auth_asym_id_list))
        df_mmCIF_auth_seq_id_list_zip = pd.DataFrame(zip(auth_seq_id_list_zip, PDB_ins_code_list))
        df_mmCIF_auth_seq_id_list_zip = df_mmCIF_auth_seq_id_list_zip.rename(columns={0: "auth_seq_id_list_zip", 1: "ins_code"})

        df_mmCIF_auth_seq_id_list_zip["PDB_with_ins_code"] = np.where(df_mmCIF_auth_seq_id_list_zip['ins_code'] != "?",
                                                                      (df_mmCIF_auth_seq_id_list_zip['auth_seq_id_list_zip'].apply(lambda x: x[0])
                                                                       + df_mmCIF_auth_seq_id_list_zip['ins_code'].apply(lambda y: y[0]) + ","
                                                                       + df_mmCIF_auth_seq_id_list_zip['auth_seq_id_list_zip'].apply(lambda x: x[1])),
                                                                      df_mmCIF_auth_seq_id_list_zip['ins_code'])

        df_mmCIF_auth_seq_id_list_zip["PDB_with_ins_code_cor"] = np.where(df_mmCIF_auth_seq_id_list_zip['PDB_with_ins_code'] != "?",
                                                                          df_mmCIF_auth_seq_id_list_zip["PDB_with_ins_code"].apply(
                                                                              lambda x: tuple(x.split(","))),
                                                                          df_mmCIF_auth_seq_id_list_zip["auth_seq_id_list_zip"])

        df_mmCIF_auth_seq_id_list_zip["auth_seq_id_list_zip"] = df_mmCIF_auth_seq_id_list_zip["PDB_with_ins_code_cor"]
        df_mmCIF_auth_seq_id_list_zip = df_mmCIF_auth_seq_id_list_zip.drop(columns=["PDB_with_ins_code_cor", "ins_code", "PDB_with_ins_code"])

        df_auth_seq_id_list_zip_final = df_mmCIF_auth_seq_id_list_zip.merge(poly_nonpoly_atom_site, left_on="auth_seq_id_list_zip",
                                                                            right_on="PDB_num_and_chain", how='left')

        df_auth_seq_id_list_zip_final["question_mark"] = np.where(
            df_auth_seq_id_list_zip_final["auth_seq_id_list_zip"].apply(lambda x: x[0] in dot_or_question_tuple),
            df_auth_seq_id_list_zip_final["auth_seq_id_list_zip"].apply(lambda x: x[0]),
            df_auth_seq_id_list_zip_final["Uni_or_50k"].apply(lambda x: x))
        try:
            df_auth_seq_id_list_zip_final["final"] = np.where(df_auth_seq_id_list_zip_final["question_mark"].apply(lambda x: type(x) == float),
                                                              df_auth_seq_id_list_zip_final["auth_seq_id_list_zip"].apply(
                                                                  lambda x: "." if x[0] == "." else
                                                                  "?" if x[0] == "?" else str(
                                                                      int(''.join(filter(str.isdigit, str(x[0])))) + default_mmCIF_num)
                                                                  if x[1] in chains_to_change else str(int(''.join(filter(str.isdigit, str(x[0])))))),
                                                              df_auth_seq_id_list_zip_final["question_mark"].apply(lambda x: x))
        except ValueError:
            # print("ValueError in table " + auth_seq_id + " has non-numeric value point in file " + mmcif_dict["data_"])
            return print("ValueError in table " + auth_seq_id + " has non-numeric value point in file " + mmcif_dict["data_"])

        df_auth_seq_id_list_zip_final["ins_code"] = df_auth_seq_id_list_zip_final["final"].apply(lambda x: "?"
        if re.sub('[0-9]+', '', x).strip("-").strip(".").strip('?') == ""
        else re.sub('[0-9]+', '', x).strip("-").strip(".").strip('?'))
        df_auth_seq_id_list_zip_final["final"] = df_auth_seq_id_list_zip_final["final"].apply(lambda x: x.strip(re.sub('[0-9\-\?\.]+', '', x)))

        for num in df_auth_seq_id_list_zip_final["final"]:
            if num == "":
                print("Empty str")
            if type(num) == float:
                print("Float or npNAN")

        # actual replacing auth_num with UniProt_num and of ins_code with '?'

        PDB_ins_code_list = list()
        if PDB_ins_code != 0:
            if "." in mmcif_dict[PDB_ins_code]:
                for ins in df_auth_seq_id_list_zip_final["ins_code"].values:
                    if "?" == ins:
                        PDB_ins_code_list.append(".")
                    else:
                        PDB_ins_code_list.append(ins)
                mmcif_dict[PDB_ins_code] = PDB_ins_code_list
            else:
                mmcif_dict[PDB_ins_code] = list(df_auth_seq_id_list_zip_final["ins_code"].values)

        if "_pdbx_branch_scheme" in auth_seq_id:
            mmcif_dict["_pdbx_branch_scheme.auth_seq_num"] = list(df_auth_seq_id_list_zip_final["final"].values)
        else:
            mmcif_dict[auth_seq_id] = list(df_auth_seq_id_list_zip_final["final"].values)

    return mmcif_dict


def column_formation(mmcif_dict):
    mmcif_dict_keys = mmcif_dict.keys()
    aut_seq_all_splitted = list()
    for key in mmcif_dict_keys:
        key_dot_splitted = key.split(".")
        for tab_name_col_name in key_dot_splitted:
            if "auth_seq" in tab_name_col_name:
                if "auth_seq_id" in key:
                    aut_seq_all_splitted.append(key_dot_splitted[:1] + key_dot_splitted[1].split("auth_seq_id"))
                if "auth_seq_num" in key:
                    aut_seq_all_splitted.append(key_dot_splitted[:1] + key_dot_splitted[1].split("auth_seq_num"))

    totaling_combinations = list()
    for table_name_prefix_suffix in aut_seq_all_splitted:
        combinations = list()
        for key in mmcif_dict_keys:
            if table_name_prefix_suffix[0] == key.split(".")[0]:
                # res_num auth_seq_id or auth_seq_num
                if table_name_prefix_suffix[1] in key and table_name_prefix_suffix[2] in key \
                        and "auth_seq_id" in key or "auth_seq_num" in key:
                    combinations.append(key)
                # chain auth_asym_id or strand_id
                if "assembly" in mmcif_dict["data_"]:
                    if table_name_prefix_suffix[1] in key and table_name_prefix_suffix[2] in key \
                            and "orig_auth_asym_id" in key:
                        combinations.append(key)
                else:
                    if table_name_prefix_suffix[1] in key and table_name_prefix_suffix[2] in key \
                            and "auth_asym_id" in key or "strand_id" in key:
                        combinations.append(key)
                # ins_code
                if table_name_prefix_suffix[1] in key and table_name_prefix_suffix[2] in key \
                        and "ins_code" in key:
                    combinations.append(key)
                # monomer_type or auth_comp_id or auth_mon_id or mon_id for _struct_ref_seq_dif
                if table_name_prefix_suffix[1] in key and table_name_prefix_suffix[2] in key \
                        and "auth_comp_id" in key or "auth_mon_id" in key:
                    combinations.append(key)
                elif table_name_prefix_suffix[0] == "_struct_ref_seq_dif" \
                        and "mon_id" in key and "db_mon_id" not in key:
                    combinations.append(key)

        # work assuming all the elements in right order
        # and they are not crossing each other
        if len(combinations) > 4:
            combinations = combinations[:4]

        ordered_combination = list()
        for name in combinations:
            if "auth_seq" in name:
                ordered_combination.insert(0, name)
        for name in combinations:
            if "auth_asym_id" in name or "strand_id" in name:
                ordered_combination.insert(1, name)
        for name in combinations:
            if "ins_code" in name:
                ordered_combination.insert(2, name)
        for name in combinations:
            if "auth_comp_id" in name or "mon_id" in name:
                ordered_combination.insert(3, name)

        # exceptions
        if (  # "pdbx_unobs_or_zero_occ_residues" not in ordered_combination[0]
                "nonpoly_scheme" not in ordered_combination[0]
                and "poly_seq_scheme" not in ordered_combination[0]
                and "ndb_struct_na_base" not in ordered_combination[0]):
            totaling_combinations.append(ordered_combination)

    return totaling_combinations


def mmCIF_parser(mmCIF_name, default_input_path_to_mmCIF, df_PDBe_PDB_UniProt_without_null_index_PDBe, default_mmCIF_num, chains_to_change,
                 chains_to_change_one_to_end):
    mmcif_dict = try_MMCIF2Dict(default_input_path_to_mmCIF, mmCIF_name)
    if mmcif_dict == 0:
        return None

    try:
        _pdbx_poly_seq_scheme_auth_seq_num_before_change = mmcif_dict["_pdbx_poly_seq_scheme.auth_seq_num"]
    except KeyError:
        _pdbx_poly_seq_scheme_auth_seq_num_before_change = mmcif_dict["_pdbe_orig_poly_seq_scheme.auth_seq_num"]
        pass

    _atom_site_label_comp_id_list = mmcif_dict["_atom_site.label_comp_id"]
    _atom_site_label_seq_id_list = mmcif_dict["_atom_site.label_seq_id"]
    _atom_site_label_asym_id = mmcif_dict["_atom_site.label_asym_id"]
    _atom_site_pdbx_PDB_ins_code = mmcif_dict["_atom_site.pdbx_PDB_ins_code"]

    _atom_site_auth_comp_id = mmcif_dict["_atom_site.auth_comp_id"]
    _atom_site_auth_seq_id = mmcif_dict["_atom_site.auth_seq_id"]
    _atom_site_auth_asym_id = mmcif_dict["_atom_site.auth_asym_id"]
    _atom_site_pdbx_formal_charge = mmcif_dict["_atom_site.pdbx_formal_charge"]

    final_mmCIF_data_list_of_tuples_just_pdb = list(zip(_atom_site_label_seq_id_list, _atom_site_label_comp_id_list, _atom_site_label_asym_id))
    final_mmCIF_data_list_of_tuples_with_auth = list(zip(_atom_site_auth_seq_id, _atom_site_auth_comp_id, _atom_site_auth_asym_id))
    final_mmCIF_data_list_of_tuples_for_df = list(
        zip(final_mmCIF_data_list_of_tuples_just_pdb, final_mmCIF_data_list_of_tuples_with_auth, _atom_site_pdbx_PDB_ins_code))

    df_mmCIF = pd.DataFrame(final_mmCIF_data_list_of_tuples_for_df)
    df_mmCIF = df_mmCIF.rename(columns={0: "One_to_N_mmCIF", 1: "auth_mmCIF", 2: "ins_code"})

    df_mmCIF["One_to_N_mmCIF"]

    df_mmCIF["PDBnum_inc_code"] = np.where(df_mmCIF['ins_code'] != "?",
                                           (df_mmCIF['auth_mmCIF'].apply(lambda x: x[0]) + df_mmCIF["ins_code"].apply(lambda y: y[0]) + ","
                                            + df_mmCIF['auth_mmCIF'].apply(lambda x: x[1]) + "," + df_mmCIF['auth_mmCIF'].apply(lambda x: x[2])),
                                           df_mmCIF["ins_code"])
    df_mmCIF["PDBnum_inc_code_cor"] = np.where(df_mmCIF["PDBnum_inc_code"] != "?", df_mmCIF["PDBnum_inc_code"].apply(lambda x: tuple(x.split(","))),
                                               df_mmCIF["auth_mmCIF"])

    df_mmCIF["auth_mmCIF"] = df_mmCIF["PDBnum_inc_code_cor"]
    df_mmCIF = df_mmCIF.drop(columns=["PDBnum_inc_code_cor", "ins_code", "PDBnum_inc_code"])

    df_PDBe_PDB_UniProt_without_null_index_PDBe = df_PDBe_PDB_UniProt_without_null_index_PDBe.reset_index()
    df_final = df_mmCIF.merge(df_PDBe_PDB_UniProt_without_null_index_PDBe, left_on="One_to_N_mmCIF", right_on="PDBe", how='left')
    df_final = df_final.rename(columns={"PDBe_copy": "PDBe"})
    df_final = df_final.drop_duplicates(subset="auth_mmCIF", keep='first')
    df_final["PDB_num_and_chain"] = df_final["auth_mmCIF"].apply(lambda x: (x[0], x[2]))
    df_final["PDBe_num_and_chain"] = df_final["One_to_N_mmCIF"].apply(lambda x: (x[0], x[2]))

    df_final["Uni_or_50k_NAN"] = np.where(df_final["One_to_N_mmCIF"].apply(lambda x: x[0] != "."),
                                          df_final["UniProt_50k"].apply(lambda x: x),
                                          df_final["PDB_num_and_chain"].apply(
                                              lambda x: str(int(''.join(filter(str.isdigit, x[0]))) + default_mmCIF_num + 10000)
                                              if x[1] in chains_to_change else str(int(''.join(filter(str.isdigit, x[0]))))))
    df_final["Uni_or_50k"] = np.where(df_final["Uni_or_50k_NAN"].apply(lambda x: type(x) == float),
                                      df_final["PDBe_num_and_chain"].apply(
                                          lambda x: "." if x[0] == "." else str(int(''.join(filter(str.isdigit, x[0]))) + default_mmCIF_num)
                                          if x[1] in chains_to_change_one_to_end else str(int(''.join(filter(str.isdigit, x[0]))))),
                                      df_final["Uni_or_50k_NAN"].apply(lambda x: x))

    df_final_atom_site = df_final[["PDBe", "PDB", "UniProt", "PDBe_num_and_chain", "PDB_num_and_chain", "AccessionID", "Uni_or_50k"]]

    return [df_final_atom_site, mmcif_dict]


def SIFTS_tree_parser(handle_SIFTS):
    tree = ET.parse(handle_SIFTS)
    root = tree.getroot()

    crossRefDb_list = list()
    PDBe_val_tuples_in_list = list()
    PDBe_val_tuples_in_list_for_Uni = list()
    PDBe_val_tuples_in_list_for_PDB = list()
    PDB_val_tuples_in_list = list()
    UniProt_val_tuple_in_list = list()
    UniProtdbAccessionId_list = list()
    UniProt_conversion_dict = dict()
    details_list = list()
    # Human_readable_AccessionID_list = list()

    for entity in root:
        if entity.tag.endswith("entity"):
            entity_chainID_list = list(entity.attrib.items())
            if entity_chainID_list[0][0] == "type" and entity_chainID_list[0][1] == "protein":
                for segment in entity:
                    for listResidue in segment:
                        if listResidue.tag.endswith("listMapRegion"):
                            for mapRegion in listResidue:
                                for db in mapRegion:
                                    dbSource_UniProt = list(db.attrib.items())
                                    if "dbSource" == dbSource_UniProt[0][0] and "UniProt" == dbSource_UniProt[0][1]:
                                        if db.text is None:
                                            UniProt = dbSource_UniProt[2][1]
                                        else:
                                            Human_readable = db.text
                                            UniProt_conversion_dict[UniProt] = Human_readable

                        for residue in listResidue:
                            key_val_tuples_in_list_parent = list(residue.attrib.items())
                            if key_val_tuples_in_list_parent[0][0] == "dbSource" and key_val_tuples_in_list_parent[0][1] == "PDBe":
                                PDBe_val_tuples_in_list.append(
                                    (key_val_tuples_in_list_parent[2][1], key_val_tuples_in_list_parent[3][1], entity_chainID_list[1][1]))

                                for crossRefDb in residue:
                                    if crossRefDb.tag.endswith("residueDetail") and crossRefDb.text != "Not_Observed":
                                        details_list.append((("PDBid", root.get("dbAccessionId")), ("Annotation:", crossRefDb.text), (
                                            key_val_tuples_in_list_parent[2][1], key_val_tuples_in_list_parent[3][1], entity_chainID_list[1][1])))

                                    crossRefDb_list.append(crossRefDb.attrib)
                                    key_val_tuples_in_list_child = list(crossRefDb.attrib.items())

                                    if key_val_tuples_in_list_child[0][0] == "dbSource" and key_val_tuples_in_list_child[0][1] == "PDB":
                                        PDB_val_tuples_in_list.append((key_val_tuples_in_list_child[3][1], key_val_tuples_in_list_child[4][1],
                                                                       key_val_tuples_in_list_child[5][1]))
                                        PDBe_val_tuples_in_list_for_PDB.append(
                                            (key_val_tuples_in_list_parent[2][1], key_val_tuples_in_list_parent[3][1], entity_chainID_list[1][1]))

                                    if key_val_tuples_in_list_child[0][0] == "dbSource" and key_val_tuples_in_list_child[0][1] == "UniProt":
                                        UniProt_val_tuple_in_list.append(
                                            (key_val_tuples_in_list_child[3][1], key_val_tuples_in_list_child[4][1], entity_chainID_list[1][1]))
                                        PDBe_val_tuples_in_list_for_Uni.append(
                                            (key_val_tuples_in_list_parent[2][1], key_val_tuples_in_list_parent[3][1], entity_chainID_list[1][1]))
                                        UniProtdbAccessionId_list.append(key_val_tuples_in_list_child[2][1])

    tuple_PDBe_for_PDB_and_tuple_PDB = list(zip(PDBe_val_tuples_in_list_for_PDB, PDB_val_tuples_in_list))
    tuple_PDBe_for_UniProt_and_tuple_UniProt = list(zip(PDBe_val_tuples_in_list_for_Uni, UniProt_val_tuple_in_list, UniProtdbAccessionId_list))

    return [tuple_PDBe_for_PDB_and_tuple_PDB, tuple_PDBe_for_UniProt_and_tuple_UniProt, UniProt_conversion_dict, details_list]


def SIFTS_data_parser_for_mmCIF(tuple_PDBe_for_PDB_and_tuple_PDB, tuple_PDBe_for_UniProt_and_tuple_UniProt, default_mmCIF_num,
                                chains_to_change="all"):
    df_PDBe_UniProt = pd.DataFrame(tuple_PDBe_for_UniProt_and_tuple_UniProt, columns=['PDBe', 'UniProt', "AccessionID"])
    df_PDBe_UniProt = df_PDBe_UniProt.drop_duplicates(subset="PDBe", keep='first')
    df_PDBe_PDB = pd.DataFrame(tuple_PDBe_for_PDB_and_tuple_PDB, columns=['PDBe', 'PDB'])
    df_PDBe_PDB = df_PDBe_PDB.drop_duplicates(subset="PDBe", keep='first')

    df_PDBe_PDB_UniProt = df_PDBe_PDB.merge(df_PDBe_UniProt, left_on="PDBe", right_on="PDBe", how='left')
    df_PDBe_PDB_UniProt['UniProt'] = df_PDBe_PDB_UniProt['UniProt'].replace(np.nan, "50000")
    df_PDBe_PDB_UniProt["Uni_moD"] = np.where(df_PDBe_PDB_UniProt['UniProt'] != "50000", df_PDBe_PDB_UniProt['UniProt'], df_PDBe_PDB_UniProt["PDBe"])
    df_PDBe_PDB_UniProt.loc[:, 'new_col_Uni'] = df_PDBe_PDB_UniProt.Uni_moD.map(lambda x: x[0])
    df_PDBe_PDB_UniProt["UniProt_50k"] = df_PDBe_PDB_UniProt.new_col_Uni.apply(lambda x: str(int(x) + default_mmCIF_num if type(x) == str else x))
    df_PDBe_PDB_UniProt.loc[df_PDBe_PDB_UniProt['UniProt'] != '50000', 'UniProt_50k'] = df_PDBe_PDB_UniProt['new_col_Uni']

    Three_Rows_CIF_Num_Uni = []
    if chains_to_change == "all":
        for index, rows in df_PDBe_PDB_UniProt.iterrows():
            intermediate_list = [rows.PDBe, rows.UniProt_50k, rows.Uni_moD, rows.PDB, rows.AccessionID]
            Three_Rows_CIF_Num_Uni.append(intermediate_list)

    else:
        for index, rows in df_PDBe_PDB_UniProt.iterrows():
            if rows.PDB[2].strip() in chains_to_change:
                intermediate_list = [rows.PDBe, rows.UniProt_50k, rows.Uni_moD, rows.PDB, rows.AccessionID]
            else:
                intermediate_list = [rows.PDBe, rows.PDB[0], rows.Uni_moD, rows.PDB, rows.AccessionID]
            Three_Rows_CIF_Num_Uni.append(intermediate_list)

    df_PDBe_PDB_UniProt["Three_Rows_CIF_Num_Uni"] = Three_Rows_CIF_Num_Uni
    df_PDBe_PDB_UniProt_without_null = df_PDBe_PDB_UniProt[df_PDBe_PDB_UniProt.PDB.map(lambda x: x[0]) != "null"]
    df_PDBe_PDB_UniProt_without_null_index_PDBe = df_PDBe_PDB_UniProt_without_null.set_index("PDBe")

    return [df_PDBe_PDB_UniProt_without_null_index_PDBe, df_PDBe_PDB_UniProt]


def master_mmCIF_renumber_function(input_mmCIF_file_were_found, default_input_path_to_mmCIF,
                                   default_input_path_to_SIFTS, default_output_path_to_mmCIF,
                                   default_mmCIF_num, gzip_mode, exception_AccessionIDs):
    input_mmCIF_assembly_files_were_found_list = list()
    input_mmCIF_assembly_files_were_found_list.append(input_mmCIF_file_were_found)

    for mmCIF_name in input_mmCIF_assembly_files_were_found_list:
        log_message = list()
        SIFTS_name = mmCIF_name[:4] + ".xml.gz"

        # for no SIFTS _no_SIFTS_out.cif.gz
        try:
            gzip.open(Path(str(default_input_path_to_SIFTS) + "/" + SIFTS_name), 'rt')
        except FileNotFoundError:
            mmcif_dict = try_MMCIF2Dict(default_input_path_to_mmCIF, mmCIF_name)
            if mmcif_dict == 0:
                continue
            copy_file(default_input_path_to_mmCIF, mmCIF_name, default_output_path_to_mmCIF, ".cif.gz", gzip_mode)
            log_message = if_no_SIFTS_data_log(mmCIF_name, mmcif_dict, log_message)
            return log_message

        # for zerobyte SIFTS _zerobyte_SIFTS_out.cif.gz
        if os.path.getsize(Path(str(default_input_path_to_SIFTS) + "/" + SIFTS_name)) == 0:
            mmcif_dict = try_MMCIF2Dict(default_input_path_to_mmCIF, mmCIF_name)
            if mmcif_dict == 0:
                continue
            copy_file(default_input_path_to_mmCIF, mmCIF_name, default_output_path_to_mmCIF, ".cif.gz", gzip_mode)
            log_message = if_no_SIFTS_data_log(mmCIF_name, mmcif_dict, log_message)
            return log_message

        product_tree_SIFTS = try_SIFTS_tree_parser(default_input_path_to_SIFTS, SIFTS_name)
        if product_tree_SIFTS == 0:
            continue

        tuple_PDBe_for_PDB_and_tuple_PDB = product_tree_SIFTS[0]
        tuple_PDBe_for_UniProt_and_tuple_UniProt = product_tree_SIFTS[1]
        UniProt_conversion_dict = product_tree_SIFTS[2]

        # _no UniProt in SIFTS _no_UniProt_in_SIFTS_out.cif.gz
        if tuple_PDBe_for_UniProt_and_tuple_UniProt == list():
            mmcif_dict = try_MMCIF2Dict(default_input_path_to_mmCIF, mmCIF_name)
            if mmcif_dict == 0:
                continue
            copy_file(default_input_path_to_mmCIF, mmCIF_name, default_output_path_to_mmCIF, ".cif.gz", gzip_mode)
            log_message = if_no_SIFTS_data_log(mmCIF_name, mmcif_dict, log_message)
            return log_message

        product_of_SIFTS_data_parser = SIFTS_data_parser_for_mmCIF(tuple_PDBe_for_PDB_and_tuple_PDB, tuple_PDBe_for_UniProt_and_tuple_UniProt,
                                                                   default_mmCIF_num, 'all')
        df_PDBe_PDB_UniProt = product_of_SIFTS_data_parser[1]

        # all good till here
        handling_chain_numbering = handling_chain_numbering_clashes(df_PDBe_PDB_UniProt, exception_AccessionIDs)
        chains_to_change = handling_chain_numbering[0]
        combined_tuple_PDBe_UniProt_AccessionID = handling_chain_numbering[1]
        longest_AccessionID_list = handling_chain_numbering[3]
        chains_to_change_one_to_end = handling_chain_numbering[4]

        product_of_SIFTS_data_parser = SIFTS_data_parser_for_mmCIF(tuple_PDBe_for_PDB_and_tuple_PDB, combined_tuple_PDBe_UniProt_AccessionID,
                                                                   default_mmCIF_num, chains_to_change)
        df_PDBe_PDB_UniProt_without_null_index_PDBe = product_of_SIFTS_data_parser[0]
        df_PDBe_PDB_UniProt = product_of_SIFTS_data_parser[1]

        renumbered_count = renumbered_count_in_chains(chains_to_change_one_to_end, df_PDBe_PDB_UniProt_without_null_index_PDBe,
                                                      mmCIF_name, UniProt_conversion_dict, longest_AccessionID_list)
        chain_total_renum = renumbered_count[0]
        nothing_changed = renumbered_count[1]

        chain_total_renum.append(nothing_changed)
        mod_log_message = chain_total_renum

        # for no change needed _no_change_out.cif.gz
        if nothing_changed == 0:
            copy_file(default_input_path_to_mmCIF, mmCIF_name, default_output_path_to_mmCIF, ".cif.gz", gzip_mode)
            return mod_log_message

        product_of_mmCIF_parser = mmCIF_parser(mmCIF_name, default_input_path_to_mmCIF, df_PDBe_PDB_UniProt_without_null_index_PDBe,
                                               default_mmCIF_num, chains_to_change, chains_to_change_one_to_end)
        df_final_atom_site = product_of_mmCIF_parser[0]
        mmcif_dict = product_of_mmCIF_parser[1]

        poly_nonpoly_append = poly_nonpoly_renum(mmcif_dict, df_PDBe_PDB_UniProt, chains_to_change, default_mmCIF_num)
        poly_nonpoly_atom_site = poly_nonpoly_append.append(df_final_atom_site).drop_duplicates(subset="PDB_num_and_chain", keep='first')

        formed_columns = column_formation(mmcif_dict)
        renumber_tables(formed_columns, mmcif_dict, poly_nonpoly_atom_site, chains_to_change, default_mmCIF_num)

        try:
            output_with_this_name_ending("_renum.cif", default_output_path_to_mmCIF, mmcif_dict, mmCIF_name=mmCIF_name,
                                         gzip_mode=gzip_mode, current_directory=current_directory)
            return mod_log_message
        except IndexError:
            # 5olg data swapped columns
            print("IndexError Warning this file is not renumbered:", mmCIF_name)
            copy_file(default_input_path_to_mmCIF, mmCIF_name, default_output_path_to_mmCIF, ".cif.gz", gzip_mode)



# -*- coding: utf-8 -*-
"""
  Copyright 2022 Mitchell Isaac Parker

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.

"""

from ..functions.color import (
    gray_hex,
    pink_hex,
    purple_hex,
    cyan_hex,
    olive_hex,
    light_blue_hex,
    light_orange_hex,
    light_olive_hex,
    light_red_hex,
    light_purple_hex,
    light_green_hex,
    light_cyan_hex,
    light_pink_hex,
    light_brown_hex,
)
from .nuc import nf_name, gdp_name, gtp_name, nf_color, gdp_color, gtp_color, nuc_class_lst
from .pharm import sp2_name, sp12_name
from .prot import gef_name, binder_name
from ..functions.coord import hb_name, no_hb_name, wmhb_name

y32_name = "Y32"
q61_name = "Q61"
y71_name = "Y71"

sw1_name = "SW1"
sw2_name = "SW2"

sw1_resids = "25-40"
sw2_resids = "56-76"

sw1_color = "#e7298a"
sw2_color = "#7570b3"

in_name = 'in'
out_name = 'out'

loop_resid_dict = {sw1_name: sw1_resids, sw2_name: sw2_resids}
loop_color_dict = {sw1_name: sw1_color, sw2_name: sw2_color}

on_name = 'On'
off_name = 'Off'

r_name = "R"
t_name = "T"

outlier_name = 'Outlier'
disorder_name = 'Disordered'

noise_name = "Noise"

outlier_color = gray_hex
disorder_color = gray_hex

y32_in_name = f"{y32_name}{in_name}"
y32_out_name = f"{y32_name}{out_name}"
y71_in_name = f"{y71_name}{in_name}"
y71_out_name = f"{y71_name}{out_name}"


y32_in_gtp_name = f"{y32_in_name}.{gtp_name}"
y32_out_gtp_name = f"{y32_out_name}.{gtp_name}"

y32_in_gdp_name = f"{y32_in_name}.{gdp_name}"
y32_out_gdp_name = f"{y32_out_name}.{gdp_name}"

y32_in_nf_name = f"{y32_in_name}.{nf_name}"
y32_out_nf_name = f"{y32_out_name}.{nf_name}"

y71_in_gtp_name = f"{y71_in_name}.{gtp_name}"
y71_out_gtp_name = f"{y71_out_name}.{gtp_name}"

y71_in_gdp_name = f"{y71_in_name}.{gdp_name}"
y71_out_gdp_name = f"{y71_out_name}.{gdp_name}"

y71_in_nf_name = f"{y71_in_name}.{nf_name}"
y71_out_nf_name = f"{y71_out_name}.{nf_name}"

sw1_gtp_in_on_name = f"{y32_in_gtp_name}-{on_name.upper()}"
sw1_gdp_out_off_name = f"{y32_out_gdp_name}-{off_name.upper()}"
sw1_nf_out_gef_name = f"{y32_out_nf_name}-{gef_name}"

sw2_gtp_in_r_name = f"{y71_in_gtp_name}-{r_name}"
sw2_gtp_in_sp12a_name = f"{y71_in_gtp_name}-{sp12_name}-A"
sw2_gtp_in_sp12b_name = f"{y71_in_gtp_name}-{sp12_name}-B"
sw2_gtp_out_t_name = f"{y71_out_gtp_name}-{t_name}"
sw2_gdp_in_sp12_name = f"{y71_in_gdp_name}-{sp12_name}"
sw2_gdp_out_sp2a_name = f"{y71_out_gdp_name}-{sp2_name}-A"
sw2_gdp_out_sp2b_name = f"{y71_out_gdp_name}-{sp2_name}-B"
sw2_gdp_out_binder_name = f"{y71_out_gdp_name}-{binder_name.upper()}"
sw2_nf_out_gef_name = f"{y71_out_nf_name}-{gef_name}"

conf_name_dict = {
    y32_in_gtp_name: {"ALBBBABBBBBABBBB": sw1_gtp_in_on_name},
    y32_out_gdp_name: {"ALBBBABBBAABBBBB": sw1_gdp_out_off_name}, 
    y32_out_nf_name: {"BBAABBBBBAABBLAA": sw1_nf_out_gef_name},
    y71_in_gtp_name: {"BBBBABAAAAAAAAAAAAABA": sw2_gtp_in_r_name,
                    "BBBBABBABAAAAAAAAAABA": sw2_gtp_in_sp12a_name,
                    "BBBBABBBBBAAAAAAAAABA": sw2_gtp_in_sp12b_name},
    y71_out_gtp_name:{"BBBBABAAAABLAAAAAAABA": sw2_gtp_out_t_name},
    y71_in_gdp_name: {"BBBAEBBABAAAAAAAAAABA": sw2_gdp_in_sp12_name},
    y71_out_gdp_name: {"BBBBEBBBABAAAAAAAAABA": sw2_gdp_out_sp2a_name,
                    "BBBALBABBBAAAAAAAAABA": sw2_gdp_out_sp2b_name,
                    "BBBBLAABBBAAAAAAAAABA": sw2_gdp_out_binder_name},
    y71_out_nf_name: {"BBABLAAABAAAAAAAAAABA": sw2_nf_out_gef_name}
}

conf_color_dict = {
    sw1_name: {
        sw1_nf_out_gef_name: nf_color,
        sw1_gdp_out_off_name: gdp_color,
        sw1_gtp_in_on_name: gtp_color,
        outlier_name: outlier_color,
        disorder_name: disorder_color,
    },
    sw2_name: {
        sw2_nf_out_gef_name: light_blue_hex,
        sw2_gdp_out_sp2a_name: light_orange_hex,
        sw2_gdp_in_sp12_name: light_olive_hex,
        sw2_gdp_out_sp2b_name: light_red_hex,
        sw2_gdp_out_binder_name: light_purple_hex,
        sw2_gtp_in_r_name: light_green_hex,
        sw2_gtp_in_sp12a_name: light_cyan_hex,
        sw2_gtp_out_t_name: light_pink_hex,
        sw2_gtp_in_sp12b_name: light_brown_hex,
        outlier_name: outlier_color,
        disorder_name: disorder_color,
    },
}

conf_nuc_color_dict = {
    sw1_name: {
        sw1_nf_out_gef_name: nf_color,
        sw1_gdp_out_off_name: gdp_color,
        sw1_gtp_in_on_name: gtp_color,
        outlier_name: outlier_color,
        disorder_name: disorder_color,
    },
    sw2_name: {
        sw2_nf_out_gef_name: nf_color,
        sw2_gdp_out_sp2a_name: gdp_color,
        sw2_gdp_in_sp12_name: gdp_color,
        sw2_gdp_out_sp2b_name: gdp_color,
        sw2_gdp_out_binder_name: gdp_color,
        sw2_gtp_in_r_name: gtp_color,
        sw2_gtp_in_sp12a_name: gtp_color,
        sw2_gtp_out_t_name: gtp_color,
        sw2_gtp_in_sp12b_name: gtp_color,
        outlier_name: outlier_color,
        disorder_name: disorder_color,
    },
}


resid_color_dict = {y32_name:{y32_in_name: pink_hex, y32_out_name: cyan_hex, disorder_name: disorder_color},
                y71_name:{y71_in_name: purple_hex, y71_out_name: olive_hex, disorder_name: disorder_color}}

hb_color_dict = {wmhb_name: cyan_hex, hb_name: purple_hex, no_hb_name: pink_hex}

y32_name_lst = list(resid_color_dict[y32_name].keys())
y71_name_lst = list(resid_color_dict[y71_name].keys())

sw1_name_lst = list(conf_color_dict[sw1_name].keys())
sw2_name_lst = list(conf_color_dict[sw2_name].keys())


# -*- coding: utf-8 -*-
"""
  Copyright 2022 Mitchell Isaac Parker

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.

"""

import urllib.request
import signal

from .path import append_file_path, delete_path, path_exists, unzip_file


def alarm_handler(signum, frame):

    raise Exception


def download_file(url, path, check=True, alarm=60, tries=3):

    append_file_path(path)

    present = False
    if check:
        present = path_exists(path)

    if not present:
        for _ in range(1, tries):
            try:
                signal.signal(signal.SIGALRM, alarm_handler)
                signal.alarm(alarm)
                urllib.request.urlretrieve(url, path)
                signal.alarm(0)
                break
            except Exception:
                delete_path(path)
                signal.alarm(0)
                alarm *= 1.5
            signal.alarm(0)


def download_unzip(url, path, check=True):

    append_file_path(path)

    present = False
    if check:
        present = path_exists(path)

    if not present:
        unzip_file(urllib.request.urlopen(url), out_path=path)



# -*- coding: utf-8 -*-
"""
  Copyright 2022 Mitchell Isaac Parker

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.

"""

import pandas as pd
import seaborn as sns
import matplotlib.pyplot as plt
from matplotlib_venn import venn2

from .color import change_hex_alpha, get_lst_colors, gray_hex, black_hex
from .path import append_file_path
from .table import (
    format_val,
    make_dict,
    lst_col,
    build_label_dict,
    mask_equal,
    get_ncols,
)
from .col import pdb_id_col

grid_hex = change_hex_alpha(gray_hex, 0.25)


def prep_plot_col(
    df,
    col,
    color_palette=None,
    rename_vals=None,
    order_lst=None,
    label_count=False,
    count_chain=True,
    count_pdb=False,
    count_cf=False,
    return_palette=True,
):

    df[col] = df[col].map(str)

    for index in list(df.index.values):

        df.at[index, col] = str(df.at[index, col]).split(" (N=")[0]

    if rename_vals is not None:
        if type(rename_vals) == dict:
            rename_dict = rename_vals
        elif type(rename_vals) == list:
            val_lst = lst_col(df, col, unique=True, return_str=True)
            rename_lst = format_val(rename_vals, return_str=True)
            rename_dict = make_dict(val_lst, rename_lst)
        df[col] = df[col].map(rename_dict)

    if order_lst is not None:
        order_lst = format_val(order_lst, return_str=True)
        df = mask_equal(df, col, order_lst)
        return_lst = order_lst.copy()
    else:
        return_lst = lst_col(df, col, unique=True, return_str=True)

    if label_count:
        col_dict = build_label_dict(
            df,
            col,
            return_str=True,
            count_chain=count_chain,
            count_pdb=count_pdb,
            count_cf=count_cf,
        )

        df[col] = df[col].map(col_dict)
        for i, row in enumerate(return_lst):
            return_lst[i] = col_dict[row]

    if type(color_palette) == dict:
        return_dict = dict()
        for val in return_lst:
            return_dict[val] = color_palette[val.split(" (")[0]]
    else:
        return_dict = color_palette

    if return_palette:
        return df, return_lst, return_dict
    else:
        return (
            df,
            return_lst,
        )


def make_legend_plot(
    plot_path,
    legend_dict,
    plot_width=2,
    plot_height=2,
    font_size=7,
    marker_shape="s",
    marker_size=3,
    legend_cols=None,
    legend_title=None,
    color_text=False,
):

    fig, ax = plt.subplots()

    fig.set_size_inches(plot_width, plot_height)

    handles = [
        plt.plot(
            [],
            [],
            marker=marker_shape,
            ls="",
            markersize=marker_size,
            markeredgewidth=0,
            markerfacecolor=color,
            label=label,
            color=color,
        )[0]
        for label, color in legend_dict.items()
    ]

    if legend_cols is None:
        legend_cols = get_ncols(list(legend_dict.keys()))

    label_color = None
    if color_text:
        label_color = "linecolor"

    legend = ax.legend(
        handles=handles,
        fontsize=font_size,
        ncol=legend_cols,
        frameon=False,
        loc="center",
        bbox_to_anchor=(0.5, 0.5),
        title=legend_title,
        labelcolor=label_color,
    )

    if legend_title is not None:
        plt.setp(legend.get_title(), fontsize=font_size)

    bbox_extra_artists = (legend,)

    plt.axis("off")

    append_file_path(plot_path)

    plt.savefig(
        plot_path,
        format="pdf",
        bbox_extra_artists=bbox_extra_artists,
        bbox_inches="tight",
        pad_inches=0.0,
        dpi=600,
    )
    plt.close()


def make_venn_plot(
    lst_1,
    lst_2,
    plot_path=None,
    label_1=None,
    label_2=None,
    color_1=None,
    color_2=None,
    color_inter=None,
    count_color="black",
    plot_title=None,
    plot_height=2,
    plot_width=2,
    font_size=7,
    alpha=0.75,
):

    if color_1 is None:
        color_1 = gray_hex
    if color_2 is None:
        color_2 = gray_hex
    if color_inter is None:
        color_inter = black_hex

    fig, ax = plt.subplots()

    fig.set_size_inches(plot_width, plot_height)

    total = len(set(lst_1 + lst_2))

    v = venn2(
        [set(lst_1), set(lst_2)],
        set_labels=(label_1, label_2),
        set_colors=(color_1, color_2),
        alpha=alpha,
        subset_label_formatter=lambda x: f"{x}\n({(x/total):1.0%})",
        ax=ax,
    )

    bbox_extra_artists = tuple()

    for text in v.set_labels:
        text.set_fontsize(font_size)
        bbox_extra_artists += (text,)

    for text in v.subset_labels:

        if text is not None:
            if text.get_text() == "0\n(0%)":
                text.set_text("")
            else:
                text.set_fontsize(font_size * 0.75)
                text.set_color(count_color)
                bbox_extra_artists += (text,)

    if plot_title is not None:
        title = fig.suptitle(plot_title, fontsize=font_size)
        bbox_extra_artists += (title,)

    if plot_path is None:
        return fig
    else:
        append_file_path(plot_path)

        plt.savefig(
            plot_path,
            format="pdf",
            bbox_extra_artists=bbox_extra_artists,
            bbox_inches="tight",
            pad_inches=0.0,
            dpi=600,
        )
        plt.close()


def make_stacked_barplot(
    plot_df,
    col_col,
    hue_col,
    plot_path,
    col_order=None,
    rename_col=None,
    col_count=False,
    hue_order=None,
    rename_hue=None,
    hue_count=False,
    hue_palette=None,
    x_str=None,
    y_str=None,
    font_size=7,
    plot_height=2,
    plot_width=2,
    line_width=0.5,
    show_legend=True,
    legend_pad=10,
    legend_cols=None,
    bar_width=0.5,
    bar_alpha=1,
    count_chain=True,
    count_pdb=False,
    count_cf=False,
    id_col=None,
    show_barh=False,
):

    df = plot_df.copy(deep=True)

    if id_col is None:
        id_column = pdb_id_col

    df, col_lst = prep_plot_col(
        df,
        col_col,
        rename_vals=rename_col,
        order_lst=col_order,
        label_count=col_count,
        count_chain=count_chain,
        count_pdb=count_pdb,
        count_cf=count_cf,
        return_palette=False,
    )

    df, hue_lst, hue_palette = prep_plot_col(
        df,
        hue_col,
        color_palette=hue_palette,
        rename_vals=rename_hue,
        order_lst=hue_order,
        label_count=hue_count,
        count_chain=count_chain,
        count_pdb=count_pdb,
        count_cf=count_cf,
    )

    hue_color_dict = get_lst_colors(hue_lst, palette=hue_palette, return_dict=True)

    sns.set_context("paper")
    sns.set_style("whitegrid")
    sns.set_palette(list(hue_color_dict.values()))

    df = pd.pivot_table(
        df,
        index=col_col,
        columns=hue_col,
        values=id_column,
        aggfunc="nunique",
    ).fillna(0)

    df = df.reindex(index=col_lst)
    df = df.reindex(columns=hue_lst)

    if show_barh:
        plot_kind = "barh"
        grid_axis = "y"
        if x_str is None:
            x_str = "% Structures"
        if y_str is None:
            y_str = col_col
    else:
        plot_kind = "bar"
        grid_axis = "x"
        if x_str is None:
            x_str = col_col
        if y_str is None:
            y_str = "% Structures"

    df.apply(lambda x: x / sum(x) * 100, axis=1).plot(
        kind=plot_kind,
        stacked=True,
        linewidth=0,
        width=bar_width,
        alpha=bar_alpha,
        figsize=(plot_width, plot_height),
        legend=show_legend,
    )

    if show_barh:
        plt.xticks(fontsize=font_size * 0.75)
    else:
        plt.xticks(fontsize=font_size * 0.75, rotation=45, ha="right")

    plt.yticks(fontsize=font_size * 0.75)
    plt.grid(axis=grid_axis, color=grid_hex, linewidth=line_width)

    x_label = plt.xlabel(x_str, fontsize=font_size)
    y_label = plt.ylabel(y_str, fontsize=font_size)

    bbox_extra_artists = (x_label, y_label)

    if show_legend or type(show_legend) == dict:

        if legend_cols is None:
            legend_cols = get_ncols(hue_lst)

        if type(show_legend) == dict:
            handles = [
                plt.plot(
                    [],
                    [],
                    marker="o",
                    ls="",
                    markeredgewidth=0,
                    markersize=3,
                    markerfacecolor=show_legend[hue],
                    label=hue,
                    color=hue,
                )[0]
                for hue in list(show_legend.keys())
            ]
        else:
            handles = [
                plt.plot(
                    [],
                    [],
                    marker="o",
                    ls="",
                    markeredgewidth=0,
                    markersize=3,
                    markerfacecolor=hue_color_dict[hue],
                    label=hue,
                )[0]
                for hue in hue_lst
            ]

        legend = plt.legend(
            handles=handles,
            fontsize=font_size * 0.75,
            ncol=legend_cols,
            loc="upper center",
            frameon=False,
            bbox_to_anchor=(0.5, 0),
            borderaxespad=legend_pad,
        )

        bbox_extra_artists += (legend,)

    sns.despine(left=True)

    plt.savefig(
        plot_path,
        format="pdf",
        bbox_inches="tight",
        bbox_extra_artists=bbox_extra_artists,
        dpi=600,
    )
    plt.close()



# -*- coding: utf-8 -*-
"""
  Copyright 2022 Mitchell Isaac Parker

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.

"""

from tqdm import tqdm
import concurrent.futures
import pymol2

from ..functions.table import merge_dicts
from ..functions.path import (
    get_file_name,
    get_interf_path,
    get_dir_path,
    delete_path,
    append_path,
    save_json,
    interf_str,
)
from ..functions.lst import (
    lst_to_str,
    sort_lst,
    lst_unique,
    type_lst,
    calc_simpson,
    calc_jaccard,
)
from ..functions.col import (
    interf_path_col,
    bound_interf_chainid_col,
    interf_area_col,
    cb_cont_col,
    atomid_cont_col,
    total_cb_cont_col,
    total_atomid_cont_col,
    iso_col,
)


byres_str = "byres"
within_str = "within"
of_str = "of"
cb_str = "and name CB"

max_cb_dist = 12.0
max_atomid_dist = 5.0


def save_interf(
    coord_path,
    interf_dir=None,
    sym_dist=5.0,
    min_area=200,
    iso_min_simi=0.7,
    use_simpson=True,
    chainid_lst=None,
):

    file_name = get_file_name(coord_path)

    if ".cif" in file_name:
        file_format = "cif"
        return_pdb = False
    elif ".pdb" in file_name:
        file_format = "pdb"
        return_pdb = True

    obj = file_name.replace(f".{file_format}", "")

    if ".gz" in file_name:
        obj = obj.replace(".gz", "")

    with pymol2.PyMOL() as pymol:

        cmd = pymol.cmd

        cmd.load(coord_path, obj)

        cmd.set("dot_solvent", 1)

        if chainid_lst is None:
            chainid_lst = cmd.get_chains(obj)
        else:
            cmd.remove(f"{obj} and not chain {lst_to_str(chainid_lst,join_txt='+')}")

        cmd.split_chains(obj)

        for chainid in chainid_lst:
            chainid_obj = f"{obj}_{chainid}"

            cmd.symexp(f"{chainid_obj}_", chainid_obj, chainid_obj, sym_dist)

        sym_obj_lst = cmd.get_object_list("all")
        sym_obj_lst.remove(obj)

        interf_dict = {coord_path: dict()}

        for chainid in chainid_lst:
            chainid_obj = f"{obj}_{chainid}"

            chainid_area = cmd.get_area(f"{chainid_obj} and polymer")

            interf = 1

            interf_dict[coord_path][chainid] = dict()

            past_cb_lst = list()
            past_atomid_lst = list()

            for sym_obj in sym_obj_lst:
                if chainid_obj != sym_obj:
                    return_chainid = cmd.get_chains(sym_obj)[0]

                    sym_chainid = f"{return_chainid}_{interf}"

                    cmd.alter(sym_obj, f"chain='{sym_chainid}'")

                    interf_obj = f"{chainid_obj}_{interf}"

                    cmd.create(interf_obj, f"{chainid_obj} {sym_obj}")

                    cmd.alter(sym_obj, f"chain='{return_chainid}'")

                    sym_area = cmd.get_area(f"{sym_obj} and polymer")
                    complex_area = cmd.get_area(f"{interf_obj} and polymer")

                    interf_area = ((chainid_area + sym_area) - complex_area) / 2

                    if interf_area >= min_area:

                        chainid_1_sele = f"{interf_obj} and chain {chainid} and polymer"
                        chainid_2_sele = (
                            f"{interf_obj} and chain {sym_chainid} and polymer"
                        )

                        chainid_cb_iterate = lst_to_str(
                            [
                                byres_str,
                                chainid_1_sele,
                                cb_str,
                                within_str,
                                max_cb_dist,
                                of_str,
                                chainid_2_sele,
                                cb_str,
                            ],
                            join_txt=" ",
                        )
                        chainid_atomid_iterate = lst_to_str(
                            [
                                byres_str,
                                chainid_1_sele,
                                within_str,
                                max_atomid_dist,
                                of_str,
                                chainid_2_sele,
                            ],
                            join_txt=" ",
                        )

                        sym_cb_iterate = lst_to_str(
                            [
                                byres_str,
                                chainid_2_sele,
                                cb_str,
                                within_str,
                                max_cb_dist,
                                of_str,
                                chainid_1_sele,
                                cb_str,
                            ],
                            join_txt=" ",
                        )

                        sym_atomid_iterate = lst_to_str(
                            [
                                byres_str,
                                chainid_2_sele,
                                within_str,
                                max_atomid_dist,
                                of_str,
                                chainid_1_sele,
                            ],
                            join_txt=" ",
                        )

                        cont_dict = {
                            "chainid_cb_lst": list(),
                            "chainid_atomid_lst": list(),
                            "sym_cb_lst": list(),
                            "sym_atomid_lst": list(),
                        }

                        cmd.iterate(
                            chainid_cb_iterate,
                            "chainid_cb_lst.append(resi)",
                            space=cont_dict,
                        )
                        cmd.iterate(
                            chainid_atomid_iterate,
                            "chainid_atomid_lst.append(resi)",
                            space=cont_dict,
                        )

                        cmd.iterate(
                            sym_cb_iterate,
                            "sym_cb_lst.append(resi)",
                            space=cont_dict,
                        )
                        cmd.iterate(
                            sym_atomid_iterate,
                            "sym_atomid_lst.append(resi)",
                            space=cont_dict,
                        )

                        chainid_cb_lst = sort_lst(
                            lst_unique(cont_dict["chainid_cb_lst"])
                        )

                        chainid_atomid_lst = sort_lst(
                            lst_unique(cont_dict["chainid_atomid_lst"])
                        )

                        sym_cb_lst = sort_lst(lst_unique(cont_dict["sym_cb_lst"]))
                        sym_atomid_lst = sort_lst(
                            lst_unique(cont_dict["sym_atomid_lst"])
                        )

                        cb_lst = sort_lst(lst_unique(chainid_cb_lst + sym_cb_lst))
                        atomid_lst = sort_lst(
                            lst_unique(chainid_atomid_lst + sym_atomid_lst)
                        )

                        if not (cb_lst in past_cb_lst or atomid_lst in past_atomid_lst):

                            past_cb_lst.append(cb_lst)
                            past_atomid_lst.append(atomid_lst)

                            total_cb_cont = len(chainid_cb_lst)
                            total_atomid_cont = len(chainid_atomid_lst)

                            if (
                                total_cb_cont >= 10 and total_atomid_cont >= 1
                            ) or total_atomid_cont >= 5:

                                interf_path = get_interf_path(
                                    obj,
                                    chainid,
                                    interf,
                                    dir_path=interf_dir,
                                    return_pdb=return_pdb,
                                )
                                cmd.save(interf_path, interf_obj)

                                iso_status = False
                                if use_simpson:
                                    cb_simi = calc_simpson(chainid_cb_lst, sym_cb_lst)
                                    atomid_simi = calc_simpson(
                                        chainid_atomid_lst, sym_atomid_lst
                                    )
                                else:
                                    cb_simi = calc_jaccard(chainid_cb_lst, sym_cb_lst)
                                    atomid_simi = calc_jaccard(
                                        chainid_atomid_lst, sym_atomid_lst
                                    )
                                if (
                                    cb_simi >= iso_min_simi
                                    and atomid_simi >= iso_min_simi
                                ):
                                    iso_status = True

                                interf_dict[coord_path][chainid][interf] = dict()

                                interf_dict[coord_path][chainid][interf][
                                    interf_path_col
                                ] = interf_path
                                interf_dict[coord_path][chainid][interf][
                                    bound_interf_chainid_col
                                ] = sym_chainid
                                interf_dict[coord_path][chainid][interf][
                                    interf_area_col
                                ] = interf_area
                                interf_dict[coord_path][chainid][interf][
                                    cb_cont_col
                                ] = lst_to_str(chainid_cb_lst)
                                interf_dict[coord_path][chainid][interf][
                                    atomid_cont_col
                                ] = lst_to_str(chainid_atomid_lst)
                                interf_dict[coord_path][chainid][interf][
                                    total_cb_cont_col
                                ] = total_cb_cont
                                interf_dict[coord_path][chainid][interf][
                                    total_atomid_cont_col
                                ] = total_atomid_cont
                                interf_dict[coord_path][chainid][interf][
                                    iso_col
                                ] = iso_status

                                interf += 1

    return interf_dict


def prep_interf(
    coord_paths,
    interf_dir=None,
    interf_json_path=None,
    sym_dist=5.0,
    min_area=200,
    iso_min_simi=0.7,
    use_simpson=True,
    chainid_dict=None,
    update_interf=False,
    num_cpu=1,
):

    interf_path = get_dir_path(dir_str=interf_str, dir_path=interf_dir)

    if update_interf:
        delete_path(interf_path)
    append_path(interf_path)

    coord_path_lst = type_lst(coord_paths)

    interf_dict = dict()

    if num_cpu == 1:
        for coord_path in tqdm(
            coord_path_lst,
            desc="Preparing interfaces",
            position=0,
            leave=True,
        ):
            interf_dict = merge_dicts(
                [
                    interf_dict,
                    save_interf(
                        coord_path,
                        interf_dir=interf_dir,
                        sym_dist=sym_dist,
                        min_area=min_area,
                        chainid_lst=chainid_dict[coord_path],
                    ),
                ]
            )
    else:
        with concurrent.futures.ProcessPoolExecutor(max_workers=num_cpu) as executor:
            job_lst = [
                executor.submit(
                    save_interf,
                    coord_path,
                    interf_dir=interf_dir,
                    sym_dist=sym_dist,
                    min_area=min_area,
                    iso_min_simi=iso_min_simi,
                    use_simpson=use_simpson,
                    chainid_lst=chainid_dict[coord_path],
                )
                for coord_path in coord_path_lst
            ]

            for job in tqdm(
                concurrent.futures.as_completed(job_lst),
                desc="Preparing interfaces",
                total=len(job_lst),
                miniters=1,
                position=0,
                leave=True,
            ):

                interf_dict = merge_dicts([interf_dict, job.result()])

    print("Prepared interfaces!")

    if interf_json_path is not None:
        save_json(interf_json_path, interf_dict)
    else:
        return interf_dict



# -*- coding: utf-8 -*-
"""
  Copyright 2022 Mitchell Isaac Parker

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.

"""

import pandas as pd
from tqdm import tqdm

from ..functions.cluster import (
    build_cutoff_dict,
    calc_dist_stat,
    order_clusters,
    build_sum_table,
)
from ..functions.table import lst_col, get_val_index_lst, build_col_count_dict
from ..functions.col import (
    cluster_col,
    total_classified_nn_col,
    total_classified_constr_col,
    nn_dist_col,
    constr_dist_col,
    nn_cutoff_col,
    constr_cutoff_col,
    total_col,
)
from ..functions.path import save_table


def classify_matrix(
    cluster_df,
    pred_df,
    fit_matrix,
    pred_matrix,
    result_table_path,
    sum_table_path=None,
    report_table_path=None,
    max_nn_dist=None,
    fit_constr_matrix=None,
    pred_constr_matrix=None,
    max_constr_dist=None,
    constr_method="mean",
    only_save_pred=False,
    reorder_class=True,
    unclass_name="Noise",
):

    if max_nn_dist is None:
        nn_cutoff_dict = build_cutoff_dict(cluster_df, fit_matrix, method="min")
    else:
        nn_cutoff_dict = dict()

    if fit_constr_matrix is not None and pred_constr_matrix is not None:
        if max_constr_dist is None:
            constr_cutoff_dict = build_cutoff_dict(
                cluster_df, fit_constr_matrix, method=constr_method
            )
        else:
            constr_cutoff_dict = dict()

    cluster_lst = lst_col(cluster_df, cluster_col, unique=True, return_str=True)

    report_dict = dict()
    for cluster in cluster_lst:
        report_dict[cluster] = {
            total_classified_nn_col: 0,
            total_classified_constr_col: 0,
        }

    for index in tqdm(
        list(pred_df.index.values),
        desc="Classifying matrix",
        position=0,
        leave=True,
    ):

        pred_lst = list()

        nn_dict = dict()
        constr_dict = dict()

        classified_dict = dict()

        for cluster in cluster_lst:

            if max_nn_dist is None:
                nn_cutoff = nn_cutoff_dict[cluster]
            else:
                nn_cutoff = max_nn_dist
                nn_cutoff_dict[cluster] = nn_cutoff

            if fit_constr_matrix is not None and pred_constr_matrix is not None:
                if max_constr_dist is None:
                    constr_cutoff = constr_cutoff_dict[cluster]
                else:
                    constr_cutoff = max_constr_dist
                    constr_cutoff_dict[cluster] = constr_cutoff

            add_cluster = True

            index_lst = get_val_index_lst(cluster_df, cluster_col, cluster)

            if len(index_lst) == 1:
                nn_dist = 0
            else:
                if len(pred_df) < 2:
                    dist_lst = pred_matrix[index_lst]
                else:
                    dist_lst = pred_matrix[index, index_lst]
                nn_dist = calc_dist_stat(
                    dist_lst,
                    method="min",
                )

            nn_dict[cluster] = nn_dist

            if nn_dist > nn_cutoff:
                add_cluster = False
            else:
                classified_dict[cluster] = total_classified_nn_col

            if (
                fit_constr_matrix is not None and pred_constr_matrix is not None
            ) and add_cluster is False:
                if len(index_lst) == 1:
                    constr_dist = 0
                else:
                    constr_dist = calc_dist_stat(
                        pred_constr_matrix[index, index_lst],
                        method=constr_method,
                    )
                constr_dict[cluster] = constr_dist

                if constr_dist > constr_cutoff:
                    add_cluster = False
                else:
                    classified_dict[cluster] = total_classified_constr_col

            if add_cluster and cluster != unclass_name:
                pred_lst.append(cluster)

        pred_cluster = unclass_name

        if len(pred_lst) == 1:
            pred_cluster = pred_lst[0]
            report_dict[pred_cluster][classified_dict[pred_cluster]] += 1

        pred_df.at[index, cluster_col] = pred_cluster

        if pred_cluster in list(nn_dict.keys()):
            nn_dist = nn_dict[pred_cluster]
        else:
            nn_dist = 0

        pred_df.at[index, nn_dist_col] = nn_dist

        if fit_constr_matrix is not None and pred_constr_matrix is not None:
            if pred_cluster in list(constr_dict.keys()):
                constr_dist = constr_dict[pred_cluster]
            else:
                constr_dist = 0
            pred_df.at[index, constr_dist_col] = constr_dist

    if not only_save_pred:
        pred_df = pd.concat([cluster_df, pred_df], sort=False)
        pred_df = pred_df.reset_index(drop=True)
        if reorder_class:
            pred_df = order_clusters(pred_df)

    save_table(result_table_path, pred_df)

    if sum_table_path is not None:
        sum_df = build_sum_table(pred_df)
        save_table(sum_table_path, sum_df)

    if report_table_path is not None:
        cluster_count_dict = build_col_count_dict(pred_df, cluster_col)
        report_df = pd.DataFrame()
        for i, cluster in enumerate(
            [x for x in cluster_lst if x in list(cluster_count_dict.keys())]
        ):
            report_df.at[i, cluster_col] = cluster
            report_df.at[i, total_col] = cluster_count_dict[cluster]
            if cluster in list(nn_cutoff_dict.keys()):
                report_df.at[i, nn_cutoff_col] = nn_cutoff_dict[cluster]
            report_df.at[i, total_classified_nn_col] = report_dict[cluster][
                total_classified_nn_col
            ]
            if fit_constr_matrix is not None and pred_constr_matrix is not None:
                if cluster in list(constr_cutoff_dict.keys()):
                    report_df.at[i, constr_cutoff_col] = constr_cutoff_dict[cluster]
                report_df.at[i, total_classified_constr_col] = report_dict[cluster][
                    total_classified_constr_col
                ]
        save_table(report_table_path, report_df)

    print("Classified matrix!")



# -*- coding: utf-8 -*-
"""
  Copyright 2022 Mitchell Isaac Parker

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.

"""

import os
import warnings

warnings.simplefilter(action="ignore", category=FutureWarning)

import argparse
import sys
import pyfiglet

from rascore.util.functions.path import (
    get_file_path,
    get_dir_name,
    copy_path,
    util_str,
    data_str,
)
from rascore.util.functions.file import entry_table_file


def main(args):

    classify = args.classify
    build = args.build
    #cluster = args.cluster
    gui = args.gui

    out = args.out
    cpu = args.cpu

    if classify is not None:

        from rascore.util.pipelines.classify_rascore import classify_rascore

        classify_rascore(
            file_paths=classify,
            out_path=out,
            num_cpu=cpu,
        )
    elif build is not None:

        from rascore.util.pipelines.prep_rascore import prep_rascore
        from rascore.util.pipelines.build_rascore import build_rascore

        pdbaa_fasta_path = None
        if build is not True:
            pdbaa_fasta_path = build
        prep_rascore(build_path=out)

        build_rascore(
            out_path=out,
            pdbaa_fasta_path=pdbaa_fasta_path,
            num_cpu=cpu,
        )

    # elif cluster is not None:
    #     from util.pipelines.prep_rascore import prep_rascore
    #     from util.pipelines.cluster_rascore import cluster_rascore

    #     prep_rascore(build_path=cluster)
    #     cluster_rascore(cluster, out_path=out, num_cpu=cpu)

    elif gui is not None:
        if gui is not True:

            from rascore.util.pipelines.prep_rascore import prep_rascore

            prep_rascore(build_path=gui)
            entry_table_path = get_file_path(entry_table_file, dir_path=gui)
            copy_path(
                entry_table_path,
                get_file_path(
                    entry_table_file,
                    dir_path=f"{get_dir_name(__file__)}/{util_str}/{data_str}",
                ),
            )

        rascore_app_path = get_file_path(
            "rascore_gui.py",
            dir_path=get_dir_name(__file__),
        )

        os.system(f"streamlit run {rascore_app_path}")


def cli(args=None):

    print(pyfiglet.figlet_format("Rascore"))
    print("A tool for analyzing RAS protein structures\n")
    print("Author: Mitchell Isaac Parker <mip34@drexel.edu>")
    print("License: Apache License 2.0\n")

    if not args:
        args = sys.argv[1:]

    parser = argparse.ArgumentParser(
        description="Rascore: A tool for analyzing RAS protein structures"
    )

    group = parser.add_mutually_exclusive_group()
    group.add_argument(
        "-classify",
        "--classify",
        nargs="+",
        required=False,
        help="path to coordinate files of RAS structures to conformationally classify provided as a space-separated list, line-sepated file, or tab-seperated file (output files saved to rascore_classify in current working directory unless an output directory path is specified)",
    )
    group.add_argument(
        "-build",
        "--build",
        type=str,
        const=True,
        nargs="?",
        required=False,
        help="build or update rascore database from the Protein Data Bank (output files saved to rascore_build in current working directory unless an output directory path is specified)",
    )
    # group.add_argument(
    #     "-cluster",
    #     "--cluster",
    #     type=str,
    #     required=False,
    #     help="path to rascore database directory (output files saved to rascore_cluster in current working directory unless an output directory path is specified)",
    # )
    group.add_argument(
        "-gui",
        "--gui",
        type=str,
        const=True,
        nargs="?",
        required=False,
        help="optional path to rascore database directory (can run without) for running the rascore GUI application",
    )
    parser.add_argument(
        "-out",
        "--out",
        type=str,
        required=False,
        help=f"output directory path (default = {os.getcwd()}",
    )
    parser.add_argument(
        "-cpu",
        "--cpu",
        type=int,
        default=1,
        required=False,
        help="number of CPUs to use (default = 1)",
    )
    args = parser.parse_args(args)
    main(args)


if __name__ == "__main__":
    cli()
